"""Sample launch module importing sklearn."""

import numpy
import scipy
import sklearn

if __name__ == "__main__":
    # The app-sklearn-import environment should NOT have access to pip or httpx
    from importlib.util import find_spec

    for disallowed in ("pip", "httpx"):
        if find_spec(disallowed):
            raise RuntimeError(f"Should not be able to import {disallowed!r}!")

    for module in (numpy, scipy, sklearn):
        # This is just here to allow the launch modules to pass lint checks
        assert module.__spec__ is not None
        assert find_spec(module.__spec__.name) is not None

    print("Environment launch module executed successfully")



"""Tests for venvstacks post-install script generation."""

import os

from pathlib import Path

from venvstacks._injected import postinstall

_EXPECTED_PYVENV_CFG = """\
home = {python_home}
include-system-site-packages = false
version = {py_version}
executable = {python_bin}
"""


def test_pyvenv_cfg() -> None:
    example_path = Path("/example/python/bin/python").absolute()
    example_version = "6.28"
    expected_pyvenv_cfg = _EXPECTED_PYVENV_CFG.format(
        python_home=str(example_path.parent),
        py_version=example_version,
        python_bin=str(example_path),
    )
    pyvenv_cfg = postinstall.generate_pyvenv_cfg(
        example_path,
        example_version,
    )
    assert pyvenv_cfg == expected_pyvenv_cfg


def test_sitecustomize_empty() -> None:
    assert postinstall.generate_sitecustomize([], []) is None


def _make_fake_paths(prefix: str, expected_line_fmt: str) -> tuple[list[Path], str]:
    # Ensure fake paths are absolute (regardless of platform)
    anchor = Path.cwd().anchor
    fake_dirs = [f"{anchor}{prefix}{n}" for n in range(5)]
    fake_paths = [Path(d) for d in fake_dirs]
    # Also report the corresponding block of expected `sitecustomize.py` lines
    expected_lines = "\n".join(expected_line_fmt.format(d) for d in fake_dirs)
    return fake_paths, expected_lines


def _make_pylib_paths() -> tuple[list[Path], str]:
    return _make_fake_paths("pylib", "addsitedir({!r})")


def _make_dynlib_paths() -> tuple[list[Path], str]:
    return _make_fake_paths("dynlib", "add_dll_directory({!r})")


def _make_missing_dynlib_paths() -> tuple[list[Path], str]:
    return _make_fake_paths("dynlib", "# Skipping {!r} (no such directory)")


def test_sitecustomize() -> None:
    pylib_paths, expected_lines = _make_pylib_paths()
    sc_text = postinstall.generate_sitecustomize(pylib_paths, [])
    assert sc_text is not None
    assert sc_text.startswith(postinstall._SITE_CUSTOMIZE_HEADER)
    assert expected_lines in sc_text
    assert "add_dll_directory(" not in sc_text
    assert "# Skipping" not in sc_text
    assert compile(sc_text, "_sitecustomize.py", "exec") is not None


def test_sitecustomize_with_dynlib() -> None:
    pylib_paths, expected_pylib_lines = _make_pylib_paths()
    dynlib_paths, expected_dynlib_lines = _make_dynlib_paths()
    sc_text = postinstall.generate_sitecustomize(
        pylib_paths, dynlib_paths, include_missing_dynlib_paths=True
    )
    assert sc_text is not None
    assert sc_text.startswith(postinstall._SITE_CUSTOMIZE_HEADER)
    assert expected_pylib_lines in sc_text
    if hasattr(os, "add_dll_directory"):
        assert expected_dynlib_lines in sc_text
    else:
        assert "add_dll_directory(" not in sc_text
        assert "# Skipping" not in sc_text
    assert compile(sc_text, "_sitecustomize.py", "exec") is not None


def test_sitecustomize_with_missing_dynlib() -> None:
    pylib_paths, expected_pylib_lines = _make_pylib_paths()
    dynlib_paths, expected_dynlib_lines = _make_missing_dynlib_paths()
    sc_text = postinstall.generate_sitecustomize(pylib_paths, dynlib_paths)
    assert sc_text is not None
    assert sc_text.startswith(postinstall._SITE_CUSTOMIZE_HEADER)
    assert expected_pylib_lines in sc_text
    if hasattr(os, "add_dll_directory"):
        assert expected_dynlib_lines in sc_text
    else:
        assert "add_dll_directory(" not in sc_text
        assert "# Skipping" not in sc_text
    assert compile(sc_text, "_sitecustomize.py", "exec") is not None



import dynlib_publisher  # Needed for os.add_dll_directory on Windows
import dynlib_consumer

if __name__ == "__main__":
    # The app environment should NOT have access to pip
    from importlib.util import find_spec

    for disallowed in ("pip",):
        if find_spec(disallowed):
            raise RuntimeError(f"Should not be able to import {disallowed!r}!")

    result = dynlib_consumer.checkdynlib_sum(1, 2)
    if result != 3:
        raise RuntimeError(f"Expected 1+2=3, got {result}")

    print("Environment launch module executed successfully")



"""Test building the sample project produces the expected results."""

import os.path
import shutil
import tempfile

from itertools import chain
from pathlib import Path
from typing import Any


# Use unittest for these test implementations due to pytest's diff handling not working
# well for these tests, as discussed in https://github.com/pytest-dev/pytest/issues/6682
from unittest import mock

import pytest

from support import (
    DeploymentTestCase,
    EnvSummary,
    LayeredEnvSummary,
    ApplicationEnvSummary,
    ManifestData,
    SpecLoadingTestCase,
    get_artifact_export_path,
    force_artifact_export,
    get_os_environ_settings,
)

from venvstacks.stacks import (
    ArchiveBuildMetadata,
    ArchiveMetadata,
    BuildEnvironment,
    StackSpec,
    LayerCategories,
)

##################################
# Sample project test helpers
##################################

_THIS_PATH = Path(__file__)
SAMPLE_PROJECT_EXPORT_DIR = _THIS_PATH.stem
SAMPLE_PROJECT_PATH = _THIS_PATH.parent / "sample_project"
SAMPLE_PROJECT_STACK_SPEC_PATH = SAMPLE_PROJECT_PATH / "venvstacks.toml"
SAMPLE_PROJECT_REQUIREMENTS_PATH = SAMPLE_PROJECT_PATH / "requirements"
SAMPLE_PROJECT_MANIFESTS_PATH = SAMPLE_PROJECT_PATH / "expected_manifests"


def _define_build_env(working_path: Path) -> BuildEnvironment:
    """Define a build environment for the sample project in a temporary folder."""
    # To simplify regeneration of committed lockfiles and metadata,
    # use the spec file directly from its checked out location
    stack_spec = StackSpec.load(SAMPLE_PROJECT_STACK_SPEC_PATH)
    build_path = working_path / "_build🐸"
    return stack_spec.define_build_environment(build_path)


def _get_expected_metadata(build_env: BuildEnvironment) -> ManifestData:
    """Path to the expected sample project archive metadata for the current platform."""
    return ManifestData(SAMPLE_PROJECT_MANIFESTS_PATH / build_env.build_platform)


def _get_expected_dry_run_result(
    build_env: BuildEnvironment, expect_tagged_outputs: bool = False
) -> dict[str, Any]:
    # Dry run results report LayerCategories instances rather than plain strings
    untagged_metadata = _get_expected_metadata(build_env).combined_data
    all_layer_manifests = untagged_metadata["layers"]
    filtered_layer_manifests: dict[LayerCategories, Any] = {}
    for category, archive_manifests in all_layer_manifests.items():
        filtered_layer_manifests[LayerCategories(category)] = archive_manifests
    # Dry run results omit any metadata keys relating solely to the built archives
    build_request_keys = (
        ArchiveBuildMetadata.__required_keys__ | ArchiveBuildMetadata.__optional_keys__
    )
    archive_keys = ArchiveMetadata.__required_keys__ | ArchiveMetadata.__optional_keys__
    archive_only_keys = archive_keys - build_request_keys
    platform_tag = build_env.build_platform
    for archive_metadata in chain(*all_layer_manifests.values()):
        for key in archive_only_keys:
            archive_metadata.pop(key, None)
        if expect_tagged_outputs:
            # Saved metadata is for untagged builds, so the tagged output dry run
            # will always indicate that a new build is needed
            # Inputs haven't changed, so the iteration number won't be increased
            install_target = archive_metadata["install_target"]
            build_iteration = archive_metadata["archive_build"]
            expected_tag = f"{platform_tag}-{build_iteration}"
            tagged_build_name = f"{install_target}-{expected_tag}"
            archive_name: str = archive_metadata["archive_name"]
            archive_suffix = archive_name.removeprefix(install_target)
            archive_metadata["archive_name"] = f"{tagged_build_name}{archive_suffix}"
    return {"layers": filtered_layer_manifests}


def _collect_locked_requirements(build_env: BuildEnvironment) -> dict[Path, str]:
    locked_requirements: dict[Path, str] = {}
    lock_dir_path = build_env.requirements_dir_path
    build_platform = build_env.build_platform
    for env in build_env.all_environments():
        env_spec = env.env_spec
        env_requirements_path = env_spec.get_requirements_path(
            build_platform, lock_dir_path
        )
        env_requirements_text = ""
        if env_requirements_path.exists():
            env_requirements_text = env_requirements_path.read_text()
        locked_requirements[env_requirements_path] = env_requirements_text
    return locked_requirements


def _export_locked_requirements(
    artifact_export_path: Path | None,
    build_env: BuildEnvironment,
    lock_paths: list[Path],
) -> None:
    if artifact_export_path is None:
        # Artifact export has not been enabled
        return
    export_dir_path = artifact_export_path / SAMPLE_PROJECT_EXPORT_DIR / "requirements"
    export_dir_path.mkdir(parents=True, exist_ok=True)
    print(f"Exporting locked requirements files to {str(export_dir_path)!r}")
    spec_dir_path = build_env.requirements_dir_path
    summary_paths = [
        p.with_name(p.name.replace("requirements-", "packages-")) for p in lock_paths
    ]
    paths_to_export = lock_paths + summary_paths
    for path_to_export in paths_to_export:
        export_path = export_dir_path / path_to_export.relative_to(spec_dir_path)
        export_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.copyfile(path_to_export, export_path)


def _export_manifests(
    manifests_export_path: Path, manifest_path: Path, archive_metadata_path: Path
) -> None:
    manifests_export_path.mkdir(parents=True, exist_ok=True)
    shutil.copyfile(manifest_path, manifests_export_path / manifest_path.name)
    shutil.copytree(
        archive_metadata_path,
        manifests_export_path / archive_metadata_path.name,
        dirs_exist_ok=True,
    )


def _export_archives(
    artifact_export_path: Path | None,
    build_env: BuildEnvironment,
    manifest_path: Path,
    archive_metadata_paths: list[Path],
    archive_paths: list[Path],
) -> None:
    print("Copying generated artifact manifest files back to source tree")
    metadata_path = SAMPLE_PROJECT_MANIFESTS_PATH / build_env.build_platform
    archive_metadata_path = Path(os.path.commonpath(archive_metadata_paths))
    _export_manifests(metadata_path, manifest_path, archive_metadata_path)
    if artifact_export_path is None:
        # Artifact export has not been enabled
        return
    # Export manifests from CI
    test_export_dir_path = artifact_export_path / SAMPLE_PROJECT_EXPORT_DIR
    export_manifests_dir_path = test_export_dir_path / "manifests"
    print(f"Exporting manifest files to {str(export_manifests_dir_path)!r}")
    _export_manifests(export_manifests_dir_path, manifest_path, archive_metadata_path)
    # Export archives from CI
    export_archives_dir_path = test_export_dir_path / "archives"
    print(f"Exporting archive files to {str(export_archives_dir_path)!r}")
    export_archives_dir_path.mkdir(parents=True, exist_ok=True)
    archive_dir_path = build_env.build_path
    for archive_path in archive_paths:
        relative_archive_path = archive_path.relative_to(archive_dir_path)
        export_archive_path = export_archives_dir_path / relative_archive_path
        export_archive_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.copyfile(archive_path, export_archive_path)


##################################
# Expected layer definitions
##################################

EXPECTED_RUNTIMES = [
    EnvSummary("cpython-3.11", ""),
    EnvSummary("cpython-3.12", ""),
]

EXPECTED_FRAMEWORKS = [
    LayeredEnvSummary("scipy", "framework-", "cpython-3.11", ()),
    LayeredEnvSummary("sklearn", "framework-", "cpython-3.12", ()),
    LayeredEnvSummary("http-client", "framework-", "cpython-3.11", ()),
]

EXPECTED_APPLICATIONS = [
    ApplicationEnvSummary("scipy-import", "app-", "cpython-3.11", ("scipy",)),
    ApplicationEnvSummary(
        "scipy-client",
        "app-",
        "cpython-3.11",
        (
            "scipy",
            "http-client",
        ),
    ),
    ApplicationEnvSummary("sklearn-import", "app-", "cpython-3.12", ("sklearn",)),
]

EXPECTED_ENVIRONMENTS = EXPECTED_RUNTIMES.copy()
EXPECTED_ENVIRONMENTS.extend(EXPECTED_FRAMEWORKS)
EXPECTED_ENVIRONMENTS.extend(EXPECTED_APPLICATIONS)

##########################
# Test cases
##########################


class TestStackSpec(SpecLoadingTestCase):
    # Test cases that only need the stack specification file

    def test_spec_loading(self) -> None:
        self.check_stack_specification(
            SAMPLE_PROJECT_STACK_SPEC_PATH,
            EXPECTED_ENVIRONMENTS,
            EXPECTED_RUNTIMES,
            EXPECTED_FRAMEWORKS,
            EXPECTED_APPLICATIONS,
        )


class TestBuildEnvironment(DeploymentTestCase):
    # Test cases that need the full build environment to exist
    EXPECTED_APP_OUTPUT = "Environment launch module executed successfully"

    working_path: Path
    build_env: BuildEnvironment

    def setUp(self) -> None:
        working_dir = tempfile.TemporaryDirectory()
        self.addCleanup(working_dir.cleanup)
        working_path = Path(working_dir.name)
        self.working_path = working_path
        self.build_env = _define_build_env(working_path)
        os_env_updates = get_os_environ_settings()
        os_env_patch = mock.patch.dict("os.environ", os_env_updates)
        os_env_patch.start()
        self.addCleanup(os_env_patch.stop)
        self.artifact_export_path = get_artifact_export_path()
        self.export_on_success = force_artifact_export()

    def test_create_environments(self) -> None:
        # Faster test to check the links between build envs are set up correctly
        # (if this fails, there's no point even trying the full slow test case)
        build_env = self.build_env
        build_env.create_environments()
        self.check_build_environments(self.build_env.all_environments())

    @pytest.mark.slow
    @pytest.mark.expected_output
    def test_build_is_reproducible(self) -> None:
        # Need long diffs to get useful output from metadata checks
        self.maxDiff = None
        # This is organised as subtests in a monolithic test sequence to reduce CI overhead
        # Separating the tests wouldn't really make them independent, unless the outputs of
        # the initial intermediate steps were checked in for use when testing the later steps.
        # Actually configuring and building the environments is executed outside the subtest
        # declarations, since actual build failures need to fail the entire test method.
        subtests_started = subtests_passed = 0  # Track subtest failures
        build_env = self.build_env
        artifact_export_path = self.artifact_export_path
        # Read expected results from committed test data
        expected_archive_metadata = _get_expected_metadata(build_env)
        expected_dry_run_result = _get_expected_dry_run_result(build_env)
        expected_tagged_dry_run_result = _get_expected_dry_run_result(
            build_env, expect_tagged_outputs=True
        )
        committed_locked_requirements = _collect_locked_requirements(build_env)
        # Create and link the layer build environments
        build_env.create_environments(lock=True)
        # Don't even try to continue if the environments aren't properly linked
        self.check_build_environments(self.build_env.all_environments())
        # Test stage: ensure lock files can be regenerated without alteration
        generated_locked_requirements = _collect_locked_requirements(build_env)
        export_locked_requirements = True
        subtests_started += 1
        with self.subTest("Ensure lock files are reproducible"):
            self.assertEqual(
                generated_locked_requirements, committed_locked_requirements
            )
            export_locked_requirements = self.export_on_success  # Only export if forced
            subtests_passed += 1
        if export_locked_requirements:
            # Lock files will already have been written back to the source tree location
            # Also export them to the CI test artifact upload path (if set)
            _export_locked_requirements(
                artifact_export_path,
                build_env,
                list(generated_locked_requirements.keys()),
            )
        # Test stage: ensure environments can be populated without building the artifacts
        build_env.create_environments()  # Use committed lock files
        subtests_started += 1
        with self.subTest("Ensure archive publication requests are reproducible"):
            # Check generation of untagged archive names
            dry_run_result = build_env.publish_artifacts(dry_run=True)[1]
            self.assertEqual(dry_run_result, expected_dry_run_result)
            # Check generation of tagged archive names
            tagged_dry_run_result = build_env.publish_artifacts(
                dry_run=True, tag_outputs=True
            )[1]
            self.assertEqual(tagged_dry_run_result, expected_tagged_dry_run_result)
            # Dry run metadata may be incorrect because the expected outputs are being updated,
            # so always continue on and execute the full archive publication subtest
            subtests_passed += 1
        subtests_started += 1
        with self.subTest(
            "Ensure dry run builds do not update lock files or manifests"
        ):
            # No changes to lock files
            post_rebuild_locked_requirements = _collect_locked_requirements(build_env)
            self.assertEqual(
                post_rebuild_locked_requirements, generated_locked_requirements
            )
            subtests_passed += 1
        # Test stage: ensure built artifacts have the expected manifest contents
        manifest_path, snippet_paths, archive_paths = build_env.publish_artifacts()
        export_published_archives = True
        subtests_started += 1
        with self.subTest("Ensure artifact metadata is reproducible"):
            # Generated metadata should match committed reference metadata
            generated_archive_metadata = ManifestData(
                manifest_path.parent, snippet_paths
            )
            self.assertEqual(
                generated_archive_metadata.combined_data,
                expected_archive_metadata.combined_data,
            )
            self.assertCountEqual(
                generated_archive_metadata.snippet_data,
                expected_archive_metadata.snippet_data,
            )
            # Archive should be emitted for every environment defined for this platform
            num_environments = len(list(build_env.all_environments()))
            self.assertEqual(len(archive_paths), num_environments)
            export_published_archives = self.export_on_success  # Only export if forced
            # No changes to lock files
            post_publish_locked_requirements = _collect_locked_requirements(build_env)
            self.assertEqual(
                post_publish_locked_requirements, generated_locked_requirements
            )
            subtests_passed += 1
        if export_published_archives:
            # Export manifests and archives to the CI test artifact upload path (if set)
            # Also write manifests back to the source tree location for local updates
            _export_archives(
                artifact_export_path,
                build_env,
                manifest_path,
                snippet_paths,
                archive_paths,
            )
        # Test stage: ensure exported environments allow launch module execution
        subtests_started += 1
        with self.subTest("Check environment export"):
            export_path = self.working_path / "_export🦎"
            export_result = build_env.export_environments(export_path)
            self.check_environment_exports(export_path, export_result)
            subtests_passed += 1

        # Work around pytest-subtests not failing the test case when subtests fail
        # https://github.com/pytest-dev/pytest-subtests/issues/76
        self.assertEqual(
            subtests_passed, subtests_started, "Fail due to failed subtest(s)"
        )

    def test_initial_operation_selection(self) -> None:
        subtests_started = subtests_passed = 0  # Track subtest failures
        build_env = self.build_env
        # Test initial state
        for env in build_env.all_environments():
            subtests_started += 1
            with self.subTest(env=env.env_name):
                self.assertIsNone(env.want_lock, "want_lock should be None")
                self.assertFalse(env.want_lock_reset, "want_lock_reset should be False")
                self.assertTrue(env.want_build, "want_build should be True")
                self.assertTrue(env.want_publish, "want_publish should be True")
                subtests_passed += 1
        self.assertEqual(
            subtests_passed, subtests_started, "Fail due to failed subtest(s)"
        )

    def test_operation_selection_default_parameters(self) -> None:
        subtests_started = subtests_passed = 0  # Track subtest failures
        build_env = self.build_env
        # Test default parameters
        build_env.select_operations()
        for env in build_env.all_environments():
            subtests_started += 1
            with self.subTest(env=env.env_name):
                self.assertFalse(env.want_lock, "want_lock should be False")
                self.assertFalse(env.want_lock_reset, "want_lock_reset should be False")
                self.assertTrue(env.want_build, "want_build should be True")
                self.assertTrue(env.want_publish, "want_publish should be True")
                subtests_passed += 1
        self.assertEqual(
            subtests_passed, subtests_started, "Fail due to failed subtest(s)"
        )

    def test_operation_selection(self) -> None:
        subtests_started = subtests_passed = 0  # Track subtest failures
        requested_operations = (
            (False, False, False),  # Don't actually do anything
            (True, False, False),  # Just lock
            (True, True, False),  # Lock and build
            (None, None, True),  # Publish (locking and building if needed)
            (False, False, True),  # Publish (without modification to current state)
            (True, True, True),  # Lock, build, and publish
        )
        build_env = self.build_env
        for requested in requested_operations:
            want_lock, want_build, want_publish = requested
            # Check lock reset flag is set regardless of whether a lock has been requested or not
            # (the reset will only take effect if the layer actually ends up getting locked)
            build_env.select_operations(
                want_lock, want_build, want_publish, reset_lock=True
            )
            for env in build_env.all_environments():
                subtests_started += 1
                with self.subTest(env=env.env_name, requested=requested):
                    self.assertEqual(env.want_lock, want_lock, "want_lock mismatch")
                    self.assertTrue(
                        env.want_lock_reset, "want_lock_reset should be True"
                    )
                    self.assertEqual(env.want_build, want_build, "want_build mismatch")
                    self.assertEqual(
                        env.want_publish, want_publish, "want_publish mismatch"
                    )
                    subtests_passed += 1
        self.assertEqual(
            subtests_passed, subtests_started, "Fail due to failed subtest(s)"
        )

    def test_filter_layers(self) -> None:
        build_env = self.build_env
        matching = ["app-*", "*-3.*", "framework-http-client"]
        expected_layers = set(
            name
            for env in build_env.all_environments()
            if not (name := env.env_name).startswith("framework-s")
        )
        self.assertNotEqual(expected_layers, set())
        self.assertEqual(build_env.filter_layers(matching), (expected_layers, set()))
        unknown = ["unknown", "app-?", "*-app"]
        unknown_set = set(unknown)
        self.assertEqual(build_env.filter_layers(unknown), (set(), unknown_set))
        combined = sorted(matching + unknown)
        self.assertEqual(
            build_env.filter_layers(combined), (expected_layers, unknown_set)
        )

    def test_layer_selection(self) -> None:
        subtests_started = subtests_passed = 0  # Track subtest failures
        included = ["framework-sklearn"]
        dependencies = ["cpython-3.12"]
        derived = ["app-sklearn-import"]
        build_env = self.build_env

        input_selection, _ = build_env.filter_layers(included)
        build_env.select_layers(input_selection, lock=True)
        for env in build_env.all_environments():
            subtests_started += 1
            env_name = env.env_name
            with self.subTest(env=env_name):
                if env_name in included:
                    self.assertTrue(
                        env.want_lock, "want_lock not set for included layer"
                    )
                    self.assertTrue(
                        env.want_build, "want_build not set for included layer"
                    )
                    self.assertTrue(
                        env.want_publish, "want_publish not set for included layer"
                    )
                elif env_name in dependencies:
                    self.assertIsNone(
                        env.want_lock, "want_lock is not None for dependency"
                    )
                    self.assertIsNone(
                        env.want_build, "want_build is not None for dependency"
                    )
                    self.assertFalse(
                        env.want_publish, "want_publish is set for dependency"
                    )
                elif env_name in derived:
                    self.assertTrue(
                        env.want_lock, "want_lock not set for derived layer"
                    )
                    self.assertTrue(
                        env.want_build, "want_build not set for derived layer"
                    )
                    self.assertTrue(
                        env.want_publish, "want_publish not set for derived layer"
                    )
                else:
                    self.assertFalse(env.want_lock, "want_lock set for excluded layer")
                    self.assertFalse(
                        env.want_build, "want_build set for excluded layer"
                    )
                    self.assertFalse(
                        env.want_publish, "want_publish set for excluded layer"
                    )
                subtests_passed += 1
        self.assertEqual(
            subtests_passed, subtests_started, "Fail due to failed subtest(s)"
        )


# TODO: Add test case for cleaning an existing build environment
# TODO: Add test case that confirms operation & layer selection has the desired effect
# TODO: Add more layer selection test cases beyond the current one (including derivation)



"""Common utilities for stack creation and venv publication."""

import os
import os.path
import re
import subprocess
import sys
import tarfile

from contextlib import contextmanager
from importlib.machinery import EXTENSION_SUFFIXES
from pathlib import Path
from typing import Any, Generator, Iterable, Literal, Mapping, Sequence, overload

WINDOWS_BUILD = hasattr(os, "add_dll_directory")

StrPath = str | os.PathLike[str]

# Set to True (either in the source or at runtime) to dump
# the full environment being passed to subprocesses
_DEBUG_SUBPROCESS_ENVS = False


def as_normalized_path(path: StrPath, /) -> Path:
    """Normalize given path and make it absolute, *without* resolving symlinks.

    Expands user directory references, but *not* environment variable references.
    """
    # Ensure user directory references are handled as absolute paths
    expanded_path = os.path.expanduser(path)
    return Path(os.path.abspath(expanded_path))


@contextmanager
def default_tarfile_filter(filter: str) -> Generator[None, None, None]:
    """Temporarily set a global tarfile filter (useful for 3rd party API warnings)."""
    if sys.version_info < (3, 12):
        # Python 3.11 or earlier, can't set a default extraction filter
        yield
        return
    # Python 3.12 or later, set a scoped default tarfile filter
    if not filter.endswith("_filter"):
        # Allow users to omit the `_filter` suffix
        filter = f"{filter}_filter"
    default_filter = getattr(tarfile, filter)
    old_filter = tarfile.TarFile.extraction_filter
    try:
        tarfile.TarFile.extraction_filter = staticmethod(default_filter)
        yield
    finally:
        tarfile.TarFile.extraction_filter = old_filter


##############################################################################
# Running Python in built/deployed/exported layer environments
##############################################################################


def get_env_python(env_path: Path) -> Path:
    """Return the main Python binary in the given Python environment."""
    if WINDOWS_BUILD:
        env_python = env_path / "Scripts" / "python.exe"
        if not env_python.exists():
            # python-build-standalone puts the Windows Python CLI
            # at the base of the runtime folder
            env_python = env_path / "python.exe"
    else:
        env_python = env_path / "bin" / "python"
    if env_python.exists():
        return env_python
    raise FileNotFoundError(f"No Python runtime found in {env_path}")


_SUBPROCESS_PYTHON_CONFIG = {
    # Ensure any Python invocations don't pick up unwanted sys.path entries
    "PYTHONNOUSERSITE": "1",
    "PYTHONSAFEPATH": "1",
    "PYTHONPATH": "",
    "PYTHONSTARTUP": "",
    # Ensure UTF-8 mode is used
    "PYTHONUTF8": "1",
    "PYTHONLEGACYWINDOWSFSENCODING": "",
    "PYTHONLEGACYWINDOWSSTDIO": "",
    # There are other dev settings that may cause problems, but are also unlikely to be set
    # See https://docs.python.org/3/using/cmdline.html#environment-variables
    # These settings were originally added to avoid the `pip-sync` issues noted
    # in https://github.com/jazzband/pip-tools/issues/2117, and then retained
    # even after the `pip-sync` dependency was removed
}


@overload
def run_python_command_unchecked(
    command: list[str],
    *,
    env: Mapping[str, str] | None = ...,
    text: Literal[True] | None = ...,
    **kwds: Any,
) -> subprocess.CompletedProcess[str]: ...
@overload
def run_python_command_unchecked(
    command: list[str],
    *,
    env: Mapping[str, str] | None = ...,
    text: Literal[False] = ...,
    **kwds: Any,
) -> subprocess.CompletedProcess[bytes]: ...
def run_python_command_unchecked(
    command: list[str],
    *,
    env: Mapping[str, str] | None = None,
    text: bool | None = True,
    **kwds: Any,
) -> subprocess.CompletedProcess[str] | subprocess.CompletedProcess[bytes]:
    # Ensure required env vars are passed down on Windows,
    # and run Python in isolated mode with UTF-8 as the text encoding
    run_env = os.environ.copy()
    # Let the target Python runtime infer whether it's part of a venv or not
    run_env.pop("VIRTUAL_ENV", None)
    if env is not None:
        run_env.update(env)
    run_env.update(_SUBPROCESS_PYTHON_CONFIG)
    if _DEBUG_SUBPROCESS_ENVS:
        import json

        print(json.dumps(run_env, indent=2, sort_keys=True))
    # Default to running in text mode,
    # but allow it to be explicitly switched off
    text = text if text else False
    encoding = "utf-8" if text else None
    result: subprocess.CompletedProcess[str] = subprocess.run(
        command, env=run_env, text=text, encoding=encoding, **kwds
    )
    return result


def run_python_command(
    # Narrow list type spec here due to the way `subprocess.run` params are typed
    command: list[str],
    **kwds: Any,
) -> subprocess.CompletedProcess[str]:
    result = run_python_command_unchecked(command, text=True, **kwds)
    result.check_returncode()
    return result


def capture_python_output(command: list[str]) -> subprocess.CompletedProcess[str]:
    return run_python_command(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)


##############################################################################
#  Finding shared libraries to link to a common location in each environment
##############################################################################


# Python 3.11/3.12 compatibility: copy implementation of fnmatch._translate
# This function is thus under the Python Software Foundation License
# rather than the license used for the rest of the package
def _translate_fnmatch(pat: str, STAR: str, QUESTION_MARK: str) -> list[str]:
    res: list[str] = []
    add = res.append
    i, n = 0, len(pat)
    while i < n:
        c = pat[i]
        i = i + 1
        if c == "*":
            # compress consecutive `*` into one
            if (not res) or res[-1] is not STAR:
                add(STAR)
        elif c == "?":
            add(QUESTION_MARK)
        elif c == "[":
            j = i
            if j < n and pat[j] == "!":
                j = j + 1
            if j < n and pat[j] == "]":
                j = j + 1
            while j < n and pat[j] != "]":
                j = j + 1
            if j >= n:
                add("\\[")
            else:
                stuff = pat[i:j]
                if "-" not in stuff:
                    stuff = stuff.replace("\\", r"\\")
                else:
                    chunks = []
                    k = i + 2 if pat[i] == "!" else i + 1
                    while True:
                        k = pat.find("-", k, j)
                        if k < 0:
                            break
                        chunks.append(pat[i:k])
                        i = k + 1
                        k = k + 3
                    chunk = pat[i:j]
                    if chunk:
                        chunks.append(chunk)
                    else:
                        chunks[-1] += "-"
                    # Remove empty ranges -- invalid in RE.
                    for k in range(len(chunks) - 1, 0, -1):
                        if chunks[k - 1][-1] > chunks[k][0]:
                            chunks[k - 1] = chunks[k - 1][:-1] + chunks[k][1:]
                            del chunks[k]
                    # Escape backslashes and hyphens for set difference (--).
                    # Hyphens that create ranges shouldn't be escaped.
                    stuff = "-".join(
                        s.replace("\\", r"\\").replace("-", r"\-") for s in chunks
                    )
                # Escape set operations (&&, ~~ and ||).
                stuff = re.sub(r"([&~|])", r"\\\1", stuff)
                i = j + 1
                if not stuff:
                    # Empty range: never match.
                    add("(?!)")
                elif stuff == "!":
                    # Negated empty range: match any character.
                    add(".")
                else:
                    if stuff[0] == "!":
                        stuff = "^" + stuff[1:]
                    elif stuff[0] in ("^", "["):
                        stuff = "\\" + stuff
                    add(f"[{stuff}]")
        else:
            add(re.escape(c))
    assert i == n
    return res


# Python 3.11/3.12 compatibility: copy implementation of glob.translate
# This function is thus under the Python Software Foundation License
# rather than the license used for the rest of the package
def _translate_glob(
    pat: str,
    *,
    recursive: bool = False,
    include_hidden: bool = False,
    seps: Sequence[str] | None = None,
) -> str:
    """Translate a pathname with shell wildcards to a regular expression.

    If `recursive` is true, the pattern segment '**' will match any number of
    path segments.

    If `include_hidden` is true, wildcards can match path segments beginning
    with a dot ('.').

    If a sequence of separator characters is given to `seps`, they will be
    used to split the pattern into segments and match path separators. If not
    given, os.path.sep and os.path.altsep (where available) are used.
    """
    if not seps:
        if os.path.altsep:
            seps = (os.path.sep, os.path.altsep)
        else:
            seps = os.path.sep
    escaped_seps = "".join(map(re.escape, seps))
    any_sep = f"[{escaped_seps}]" if len(seps) > 1 else escaped_seps
    not_sep = f"[^{escaped_seps}]"
    if include_hidden:
        one_last_segment = f"{not_sep}+"
        one_segment = f"{one_last_segment}{any_sep}"
        any_segments = f"(?:.+{any_sep})?"
        any_last_segments = ".*"
    else:
        one_last_segment = f"[^{escaped_seps}.]{not_sep}*"
        one_segment = f"{one_last_segment}{any_sep}"
        any_segments = f"(?:{one_segment})*"
        any_last_segments = f"{any_segments}(?:{one_last_segment})?"

    results = []
    parts = re.split(any_sep, pat)
    last_part_idx = len(parts) - 1
    for idx, part in enumerate(parts):
        if part == "*":
            results.append(one_segment if idx < last_part_idx else one_last_segment)
        elif recursive and part == "**":
            if idx < last_part_idx:
                if parts[idx + 1] != "**":
                    results.append(any_segments)
            else:
                results.append(any_last_segments)
        else:
            if part:
                if not include_hidden and part[0] in "*?":
                    results.append(r"(?!\.)")
                results.extend(_translate_fnmatch(part, f"{not_sep}*", not_sep))
            if idx < last_part_idx:
                results.append(any_sep)
    res = "".join(results)
    return rf"(?s:{res})\Z"


def _ext_to_suffixes(extension: str) -> tuple["str", ...]:
    suffix_parts = extension.split(".")
    return tuple(f".{part}" for part in suffix_parts if part)


_PYLIB_SUFFIX = ".so"  # .dylib is never importable as a Python module, even on macOS
_LIB_SUFFIXES = frozenset((_PYLIB_SUFFIX, ".dylib"))

# Skip libraries with extensions that are explicitly for importable Python extension modules
_IGNORED_SUFFIXES = frozenset(
    _ext_to_suffixes(ext) for ext in EXTENSION_SUFFIXES if ext != _PYLIB_SUFFIX
)


def find_shared_libraries(
    py_version: tuple[str, str], base_path: Path, *, excluded: Iterable[str] = ()
) -> Generator[Path, None, None]:
    """Find non-extension-module shared libraries in specified directory."""
    exclusion_patterns = [
        re.compile(_translate_glob(f"/{trailing}", recursive=True))
        for trailing in excluded
    ]
    py_version_nodot = "".join(py_version)
    running_py_version_nodot = "".join(map(str, sys.version_info[:2]))
    if py_version_nodot == running_py_version_nodot:
        ignored_suffixes = _IGNORED_SUFFIXES
    else:
        ignored_suffixes = frozenset(
            _ext_to_suffixes(ext.replace(running_py_version_nodot, py_version_nodot))
            for ext in EXTENSION_SUFFIXES
            if ext != _PYLIB_SUFFIX
        )

    # Python 3.11 compatibility: use os.walk instead of Path.walk
    for this_dir, _, files in os.walk(base_path):
        dir_path = Path(this_dir)
        for fname in files:
            file_path = dir_path / fname
            if file_path.suffix not in _LIB_SUFFIXES:
                continue
            if tuple(file_path.suffixes) in ignored_suffixes:
                continue
            if any(p.search(str(file_path)) for p in exclusion_patterns):
                continue
            yield file_path


def map_symlink_targets(
    symlink_dir_path: Path, target_paths: Iterable[Path]
) -> tuple[dict[Path, Path], dict[Path, set[Path]]]:
    targets_to_link: dict[Path, Path] = {}
    ambiguous_link_targets: dict[Path, set[Path]] = {}
    for target_path in target_paths:
        symlink_path = symlink_dir_path / target_path.name
        if symlink_path in ambiguous_link_targets:
            # Already ambiguous, record another potential target
            ambiguous_link_targets[symlink_path].add(target_path)
            continue
        existing_path = targets_to_link.get(symlink_path, None)
        if existing_path is not None:
            # This name already has a target
            if existing_path == target_path:
                continue
            # Mark the link as ambiguous
            existing_path = targets_to_link.pop(symlink_path)
            ambiguous_link_targets[symlink_path] = {existing_path, target_path}
            continue
        targets_to_link[symlink_path] = target_path
    return targets_to_link, ambiguous_link_targets



"""Test support for venvstacks testing."""

import json
import os
import subprocess
import sys
import tomllib
import unittest

from dataclasses import dataclass, fields
from pathlib import Path
from typing import Any, Callable, cast, Iterable, Mapping, Sequence, TypeVar
from unittest.mock import create_autospec

import pytest

from venvstacks._util import get_env_python, capture_python_output
from venvstacks._injected.postinstall import DEPLOYED_LAYER_CONFIG

from venvstacks.stacks import (
    BuildEnvironment,
    EnvNameDeploy,
    ExportedEnvironmentPaths,
    ExportMetadata,
    LayerBaseName,
    LayerEnvBase,
    LayerVariants,
    PackageIndexConfig,
    StackSpec,
)

_THIS_DIR = Path(__file__).parent

##################################
# Marking test cases
##################################

# Basic marking uses the pytest.mark API directly
# See pyproject.toml and tests/README.md for the defined marks


def requires_venv(description: str) -> pytest.MarkDecorator:
    """Skip test case when running tests outside a virtual environment."""
    return pytest.mark.skipif(
        sys.prefix == sys.base_prefix,
        reason=f"{description} requires test execution in venv",
    )


##################################
# Exporting test artifacts
##################################

TEST_EXPORT_ENV_VAR = (
    "VENVSTACKS_EXPORT_TEST_ARTIFACTS"  # Output directory for artifacts
)
FORCED_EXPORT_ENV_VAR = "VENVSTACKS_FORCE_TEST_EXPORT"  # Force export if non-empty


def get_artifact_export_path() -> Path | None:
    """Location to export notable artifacts generated during test execution."""
    export_dir = os.environ.get(TEST_EXPORT_ENV_VAR)
    if not export_dir:
        return None
    export_path = Path(export_dir)
    if not export_path.exists():
        return None
    return export_path


def force_artifact_export() -> bool:
    """Indicate artifacts should be exported even if a test case passes."""
    # Export is forced if the environment var is defined and non-empty
    return bool(os.environ.get(FORCED_EXPORT_ENV_VAR))


####################################
# Ensuring predictable test output
####################################

# Note: tests that rely on the expected output config should be
#       marked as "expected_output" tests so they're executed
#       when regenerating the expected output files

_OUTPUT_CONFIG_PATH = _THIS_DIR / "expected-output-config.toml"
_OUTPUT_CONFIG: Mapping[str, Any] | None = None


def _cast_config(config_mapping: Any) -> Mapping[str, str]:
    return cast(Mapping[str, str], config_mapping)


def get_output_config() -> Mapping[str, Any]:
    global _OUTPUT_CONFIG
    if _OUTPUT_CONFIG is None:
        data = _OUTPUT_CONFIG_PATH.read_text()
        _OUTPUT_CONFIG = tomllib.loads(data)
    return _OUTPUT_CONFIG


def get_pinned_dev_packages() -> Mapping[str, str]:
    return _cast_config(get_output_config()["pinned-dev-packages"])


def get_os_environ_settings() -> Mapping[str, str]:
    return _cast_config(get_output_config()["env"])


##################################
# Expected layer definitions
##################################


# Runtimes
@dataclass(frozen=True)
class EnvSummary:
    _spec_name: str
    env_prefix: str

    @property
    def spec_name(self) -> LayerBaseName:
        return LayerBaseName(self._spec_name)

    @property
    def env_name(self) -> EnvNameDeploy:
        return EnvNameDeploy(self.env_prefix + self._spec_name)


# Frameworks
@dataclass(frozen=True)
class LayeredEnvSummary(EnvSummary):
    runtime_spec_name: str
    framework_spec_names: tuple[str, ...]


# Applications
@dataclass(frozen=True)
class ApplicationEnvSummary(LayeredEnvSummary):
    pass


############################################
# Reading published and exported manifests
############################################


class ManifestData:
    # Speculative: should this helper class be part of the public venvstacks API?
    combined_data: dict[str, Any]
    snippet_data: list[dict[str, Any]]

    def __init__(self, metadata_path: Path, snippet_paths: list[Path] | None = None):
        if metadata_path.suffix == ".json":
            manifest_path = metadata_path
            metadata_path = metadata_path.parent
        else:
            manifest_path = metadata_path / BuildEnvironment.METADATA_MANIFEST
        if manifest_path.exists():
            manifest_data = json.loads(manifest_path.read_text("utf-8"))
            if not isinstance(manifest_data, dict):
                msg = f"{manifest_path!r} data is not a dict: {manifest_data!r}"
                raise TypeError(msg)
            self.combined_data = manifest_data
        else:
            self.combined_data = {}
        self.snippet_data = snippet_data = []
        if snippet_paths is None:
            snippet_base_path = metadata_path / BuildEnvironment.METADATA_ENV_DIR
            if snippet_base_path.exists():
                snippet_paths = sorted(snippet_base_path.iterdir())
            else:
                snippet_paths = []
        for snippet_path in snippet_paths:
            metadata_snippet = json.loads(snippet_path.read_text("utf-8"))
            if not isinstance(metadata_snippet, dict):
                msg = f"{snippet_path!r} data is not a dict: {metadata_snippet!r}"
                raise TypeError(msg)
            snippet_data.append(metadata_snippet)


##################################
# Expected package index access
##################################


def make_mock_index_config(reference_config: PackageIndexConfig | None = None) -> Any:
    if reference_config is None:
        reference_config = PackageIndexConfig()
    mock_config = create_autospec(reference_config, spec_set=True)
    # Make conditional checks and iteration reflect the actual field values
    checked_methods = ("__bool__", "__len__", "__iter__")
    for field in fields(reference_config):
        attr_name = field.name
        mock_field = getattr(mock_config, attr_name)
        field_value = getattr(reference_config, attr_name)
        for method_name in checked_methods:
            mock_method = getattr(mock_field, method_name, None)
            if mock_method is None:
                continue
            mock_method.side_effect = getattr(field_value, method_name)
    # Still call the actual CLI arg retrieval methods
    for attr_name in dir(reference_config):
        if not attr_name.startswith(("_get_pip_", "_get_uv_")):
            continue
        mock_method = getattr(mock_config, attr_name)
        mock_method.side_effect = getattr(reference_config, attr_name)
    return mock_config


##############################################
# Running commands in a deployed environment
##############################################


def get_sys_path(env_python: Path) -> list[str]:
    command = [
        str(env_python),
        "-X",
        "utf8",
        "-Ic",
        "import json, sys; print(json.dumps(sys.path))",
    ]
    result = capture_python_output(command)
    return cast(list[str], json.loads(result.stdout))


def run_module(env_python: Path, module_name: str) -> subprocess.CompletedProcess[str]:
    command = [str(env_python), "-X", "utf8", "-Im", module_name]
    return capture_python_output(command)


#######################################################
# Checking specification loading for expected details
#######################################################


class SpecLoadingTestCase(unittest.TestCase):
    """Native unittest test case with additional spec loading validation checks."""

    def check_stack_specification(
        self,
        expected_spec_path: Path,
        expected_environments: Sequence[EnvSummary],
        expected_runtimes: Sequence[EnvSummary],
        expected_frameworks: Sequence[LayeredEnvSummary],
        expected_applications: Sequence[LayeredEnvSummary],
    ) -> None:
        stack_spec = StackSpec.load(expected_spec_path)
        runtime_keys = list(stack_spec.runtimes)
        framework_keys = list(stack_spec.frameworks)
        application_keys = list(stack_spec.applications)
        spec_keys = runtime_keys + framework_keys + application_keys
        self.assertCountEqual(spec_keys, set(spec_keys))
        expected_spec_names = [env.spec_name for env in expected_environments]
        self.assertCountEqual(spec_keys, expected_spec_names)
        spec_names = [env.name for env in stack_spec.all_environment_specs()]
        self.assertCountEqual(spec_names, expected_spec_names)
        expected_env_names = [env.env_name for env in expected_environments]
        env_names = [env.env_name for env in stack_spec.all_environment_specs()]
        self.assertCountEqual(env_names, expected_env_names)
        for rt_summary in expected_runtimes:
            spec_name = rt_summary.spec_name
            rt_env = stack_spec.runtimes[spec_name]
            self.assertEqual(rt_env.name, spec_name)
            self.assertEqual(rt_env.env_name, rt_summary.env_name)
        del spec_name, rt_env, rt_summary
        for fw_summary in expected_frameworks:
            spec_name = fw_summary.spec_name
            fw_env = stack_spec.frameworks[spec_name]
            self.assertEqual(fw_env.name, spec_name)
            self.assertEqual(fw_env.env_name, fw_summary.env_name)
            self.assertEqual(fw_env.runtime.name, fw_summary.runtime_spec_name)
            fw_dep_names = tuple(spec.name for spec in fw_env.frameworks)
            self.assertEqual(fw_dep_names, fw_summary.framework_spec_names)
        del spec_name, fw_dep_names, fw_env, fw_summary
        for app_summary in expected_applications:
            spec_name = app_summary.spec_name
            app_env = stack_spec.applications[spec_name]
            self.assertEqual(app_env.name, spec_name)
            self.assertEqual(app_env.env_name, app_summary.env_name)
            self.assertEqual(app_env.runtime.name, app_summary.runtime_spec_name)
            fw_dep_names = tuple(spec.name for spec in app_env.frameworks)
            self.assertEqual(fw_dep_names, app_summary.framework_spec_names)
        del spec_name, fw_dep_names, app_env, app_summary
        # Check path attributes
        self.assertEqual(stack_spec.spec_path, expected_spec_path)
        expected_requirements_dir_path = expected_spec_path.parent / "requirements"
        self.assertEqual(
            stack_spec.requirements_dir_path, expected_requirements_dir_path
        )


#######################################################
# Checking deployed environments for expected details
#######################################################


_T = TypeVar("_T", bound=Mapping[str, Any])


class DeploymentTestCase(unittest.TestCase):
    """Native unittest test case with additional deployment validation checks."""

    EXPECTED_APP_OUTPUT = ""

    def assertPathExists(self, expected_path: Path) -> None:
        self.assertTrue(expected_path.exists(), f"No such path: {str(expected_path)}")

    def assertPathContains(self, containing_path: Path, contained_path: Path) -> None:
        self.assertTrue(
            contained_path.is_relative_to(containing_path),
            f"{str(containing_path)!r} is not a parent folder of {str(contained_path)!r}",
        )

    def assertSysPathEntry(self, expected: str, env_sys_path: Sequence[str]) -> None:
        self.assertTrue(
            any(expected in path_entry for path_entry in env_sys_path),
            f"No entry containing {expected!r} found in {env_sys_path}",
        )

    def check_env_sys_path(
        self,
        env_path: Path,
        env_sys_path: Sequence[str],
        *,
        self_contained: bool = False,
    ) -> None:
        sys_path_entries = [Path(path_entry) for path_entry in env_sys_path]
        # Regardless of env type, sys.path entries must be absolute
        self.assertTrue(
            all(p.is_absolute() for p in sys_path_entries),
            f"Relative path entry found in {env_sys_path}",
        )
        # Regardless of env type, sys.path entries must exist
        # (except the stdlib's optional zip archive entry)
        for path_entry in sys_path_entries:
            if path_entry.suffix:
                continue
            self.assertPathExists(path_entry)

        # Check for sys.path references outside this environment
        def _is_relative_to(p: Path, base_path: Path) -> bool:
            # Also accept paths which have been fully resolved by the Python runtime
            return p.is_relative_to(base_path) or p.is_relative_to(base_path.resolve())

        if self_contained:
            # All sys.path entries should be inside the environment
            self.assertTrue(
                all(_is_relative_to(p, env_path) for p in sys_path_entries),
                f"Path outside deployed {env_path} in {env_sys_path}",
            )
        else:
            # All sys.path entries should be inside the environment's parent,
            # but at least one sys.path entry should refer to a peer environment
            peer_env_path = env_path.parent
            self.assertTrue(
                all(_is_relative_to(p, peer_env_path) for p in sys_path_entries),
                f"Path outside deployed {peer_env_path} in {env_sys_path}",
            )
            self.assertFalse(
                all(_is_relative_to(p, env_path) for p in sys_path_entries),
                f"No path outside deployed {env_path} in {env_sys_path}",
            )

    def check_build_environments(self, build_envs: Iterable[LayerEnvBase]) -> None:
        for env in build_envs:
            env_path = env.env_path
            config_path = env_path / DEPLOYED_LAYER_CONFIG
            self.assertPathExists(config_path)
            layer_config = json.loads(config_path.read_text(encoding="utf-8"))
            env_python = env_path / layer_config["python"]
            expected_python_path = env.python_path
            self.assertEqual(str(env_python), str(expected_python_path))
            base_python_path = env_path / layer_config["base_python"]
            is_runtime_env = env.kind == LayerVariants.RUNTIME
            if is_runtime_env:
                # base_python should refer to the runtime layer itself
                expected_base_python_path = expected_python_path
            else:
                # base_python should refer to the venv's base Python runtime
                self.assertIsNotNone(env.base_python_path)
                assert env.base_python_path is not None
                base_python_path = Path(os.path.normpath(base_python_path))
                expected_base_python_path = env.base_python_path
            self.assertEqual(str(base_python_path), str(expected_base_python_path))
            env_sys_path = get_sys_path(env_python)
            # Base runtime environments are expected to be self-contained
            self.check_env_sys_path(
                env_path, env_sys_path, self_contained=is_runtime_env
            )

    def check_deployed_environments(
        self,
        layered_metadata: dict[str, Sequence[_T]],
        get_env_details: Callable[[_T], tuple[str, Path, list[str]]],
    ) -> None:
        for rt_env in layered_metadata["runtimes"]:
            env_name, env_path, env_sys_path = get_env_details(rt_env)
            self.assertTrue(env_sys_path)  # Environment should have sys.path entries
            # Runtime environment layer should be completely self-contained
            self.check_env_sys_path(env_path, env_sys_path, self_contained=True)
        for fw_env in layered_metadata["frameworks"]:
            env_name, env_path, env_sys_path = get_env_details(fw_env)
            self.assertTrue(env_sys_path)  # Environment should have sys.path entries
            # Frameworks are expected to reference *at least* their base runtime environment
            self.check_env_sys_path(env_path, env_sys_path)
            # Framework and runtime should both appear in sys.path
            runtime_layer = fw_env["runtime_layer"]
            short_runtime_name = ".".join(runtime_layer.split(".")[:2])
            self.assertSysPathEntry(env_name, env_sys_path)
            self.assertSysPathEntry(short_runtime_name, env_sys_path)
        for app_env in layered_metadata["applications"]:
            env_name, env_path, env_sys_path = get_env_details(app_env)
            self.assertTrue(env_sys_path)  # Environment should have sys.path entries
            # Applications are expected to reference *at least* their base runtime environment
            self.check_env_sys_path(env_path, env_sys_path)
            # Application, frameworks and runtime should all appear in sys.path
            runtime_layer = app_env["runtime_layer"]
            short_runtime_name = ".".join(runtime_layer.split(".")[:2])
            self.assertSysPathEntry(env_name, env_sys_path)
            self.assertTrue(
                any(env_name in path_entry for path_entry in env_sys_path),
                f"No entry containing {env_name} found in {env_sys_path}",
            )
            for fw_env_name in app_env["required_layers"]:
                self.assertSysPathEntry(fw_env_name, env_sys_path)
            self.assertSysPathEntry(short_runtime_name, env_sys_path)
            # Launch module should be executable
            env_config_path = env_path / DEPLOYED_LAYER_CONFIG
            env_config = json.loads(env_config_path.read_text(encoding="utf-8"))
            env_python = env_path / env_config["python"]
            launch_module = app_env["app_launch_module"]
            try:
                launch_result = run_module(env_python, launch_module)
            except subprocess.CalledProcessError as exc:
                print(exc)
                print(exc.stdout)
                print(exc.stderr)
                raise
            # Tolerate extra trailing whitespace on stdout
            self.assertEqual(launch_result.stdout.rstrip(), self.EXPECTED_APP_OUTPUT)
            # Nothing at all should be emitted on stderr
            self.assertEqual(launch_result.stderr, "")

    def check_environment_exports(
        self, export_path: Path, export_paths: ExportedEnvironmentPaths
    ) -> None:
        metadata_path, snippet_paths, env_paths = export_paths
        exported_manifests = ManifestData(metadata_path, snippet_paths)
        env_name_to_path: dict[str, Path] = {}
        for env_metadata, env_path in zip(exported_manifests.snippet_data, env_paths):
            # TODO: Check more details regarding expected metadata contents
            self.assertPathExists(env_path)
            self.assertPathContains(export_path, env_path)
            env_name = EnvNameDeploy(env_metadata["install_target"])
            self.assertEqual(env_path.name, env_name)
            env_name_to_path[env_name] = env_path
        layered_metadata = exported_manifests.combined_data["layers"]

        def get_exported_env_details(
            env: ExportMetadata,
        ) -> tuple[EnvNameDeploy, Path, list[str]]:
            env_name = env["install_target"]
            env_path = env_name_to_path[env_name]
            env_python = get_env_python(env_path)
            env_sys_path = get_sys_path(env_python)
            return env_name, env_path, env_sys_path

        self.check_deployed_environments(layered_metadata, get_exported_env_details)



"""Sample CLI helper module importing scipy and httpx."""

import numpy
import scipy
import httpx


def main():
    # The app-scipy-client environment should NOT have access to pip, or sklearn
    from importlib.util import find_spec

    for disallowed in ("pip", "sklearn"):
        if find_spec(disallowed):
            raise RuntimeError(f"Should not be able to import {disallowed!r}!")

    for module in (numpy, scipy, httpx):
        # This is just here to allow the launch modules to pass lint checks
        assert module.__spec__ is not None
        assert find_spec(module.__spec__.name) is not None

    print("Environment launch module executed successfully")



"""venvstacks layer post-installation script.

* Loads `./share/venv/metadata/venvstacks_layer.json`
* Generates `pyvenv.cfg` for layered environments
* Generates `sitecustomize.py` for layered environments
* Precompiles all Python files in the library folder

This post-installation script is automatically injected when packing environments.
"""

import json
import os

from compileall import compile_dir
from os.path import abspath
from pathlib import Path
from typing import cast, Iterable, NotRequired, Sequence, TypedDict

DEPLOYED_LAYER_CONFIG = "share/venv/metadata/venvstacks_layer.json"


class LayerConfig(TypedDict):
    """Additional details needed to fully configure deployed environments."""

    # fmt: off
    python: str                      # Relative path to this layer's Python executable
    py_version: str                  # Expected X.Y.Z Python version for this environment
    base_python: str                 # Relative path from layer dir to base Python executable
    site_dir: str                    # Relative path to site-packages within this layer
    pylib_dirs: Sequence[str]        # Relative paths to additional sys.path entries
    dynlib_dirs: Sequence[str]       # Relative paths to additional shared library directories
    launch_module: NotRequired[str]  # Module to run with `-m` to launch the application
    # fmt: on

    # All relative paths are relative to the layer folder (and may refer to peer folders)
    # Base runtime layers will have "python" and "base_python" set to the same value
    # Application layers will have "launch_module" set


class ResolvedLayerConfig(TypedDict):
    """LayerConfig with relative paths resolved for a specific layer location."""

    # fmt: off
    layer_path: Path                 # Absolute path to layer environment
    python_path: Path                # Absolute path to this layer's Python executable
    py_version: str                  # Expected X.Y.Z Python version for this environment
    base_python_path: Path           # Absolute path from layer dir to base Python executable
    site_path: Path                  # Absolute path to site-packages within this layer
    pylib_paths: Sequence[Path]      # Absolute paths to additional sys.path entries
    dynlib_paths: Sequence[Path]     # Absolute paths to additional shared library directories
    launch_module: str|None          # Module to run with `-m` to launch the application
    # fmt: on


def load_layer_config(layer_path: Path) -> ResolvedLayerConfig:
    """Read and resolve config for the specified layer environment."""

    def deployed_path(relative_path: str) -> Path:
        """Normalize path and make it absolute, *without* resolving symlinks."""
        absolute_path = Path(abspath(layer_path / relative_path))
        if not absolute_path.is_relative_to(layer_path.parent):
            err_msg = f"Layer path ({str(absolute_path)!r} is not inside {str(layer_path.parent)!r})"
            raise RuntimeError(err_msg)
        return absolute_path

    config_path = layer_path / DEPLOYED_LAYER_CONFIG
    config_text = config_path.read_text(encoding="utf-8")
    # Tolerate runtime errors for incorrectly generated config files
    config = cast(LayerConfig, json.loads(config_text))
    return ResolvedLayerConfig(
        layer_path=layer_path,
        python_path=deployed_path(config["python"]),
        py_version=config["py_version"],
        base_python_path=deployed_path(config["base_python"]),
        site_path=deployed_path(config["site_dir"]),
        pylib_paths=[deployed_path(d) for d in config["pylib_dirs"]],
        dynlib_paths=[deployed_path(d) for d in config["dynlib_dirs"]],
        launch_module=config.get("launch_module", None),
    )


def check_absolute_path(env_path: Path) -> None:
    """Ensures given path is absolute (raises RuntimeError otherwise)."""
    if not env_path.is_absolute():
        err_msg = f"Post-installation must use absolute environment paths ({str(env_path)!r} is relative)"
        raise RuntimeError(err_msg)


def check_absolute_paths(env_paths: Iterable[Path]) -> None:
    """Ensures all given paths are absolute (raises RuntimeError otherwise)."""
    for env_path in env_paths:
        check_absolute_path(env_path)


def generate_pyvenv_cfg(base_python_path: Path, py_version: str) -> str:
    """Generate `pyvenv.cfg` contents for given base Python path and version."""
    check_absolute_path(base_python_path)
    venv_config_lines = [
        f"home = {base_python_path.parent}",
        "include-system-site-packages = false",
        f"version = {py_version}",
        f"executable = {base_python_path}",
        "",
    ]
    return "\n".join(venv_config_lines)


_SITE_CUSTOMIZE_HEADER = '''\
"""venvstacks layered environment site customization script

* Calls `site.addsitedir` for any configured Python path entries
* Calls `os.add_dll_directory` for any configured Windows dynlib paths

This venv module is automatically generated by the post-installation script.
"""

'''


def generate_sitecustomize(
    pylib_paths: Sequence[Path],
    dynlib_paths: Sequence[Path],
    *,
    include_missing_dynlib_paths: bool = False,
) -> str | None:
    """Generate `sitecustomize.py` contents for given linked environment directories."""
    sc_lines = [_SITE_CUSTOMIZE_HEADER]
    if pylib_paths:
        check_absolute_paths(pylib_paths)
        pylib_contents = [
            "# Allow loading modules and packages from linked environments",
            "from site import addsitedir",
        ]
        for path_entry in pylib_paths:
            pylib_contents.append(f"addsitedir({str(path_entry)!r})")
        pylib_contents.append("")
        sc_lines.extend(pylib_contents)
    if dynlib_paths and hasattr(os, "add_dll_directory"):
        # Python packages for Windows generally handle their own dynamic import config
        # This works around the cases that don't
        check_absolute_paths(dynlib_paths)
        dynlib_contents = [
            "# Allow loading misplaced DLLs on Windows",
            "from os import add_dll_directory",
        ]
        for dynlib_path in dynlib_paths:
            if include_missing_dynlib_paths or dynlib_path.exists():
                dynlib_entry = f"add_dll_directory({str(dynlib_path)!r})"
            else:
                # Nothing added DLLs to this folder at build time, so skip it
                # (add_dll_directory fails if the specified folder doesn't exist)
                dynlib_entry = f"# Skipping {str(dynlib_path)!r} (no such directory)"
            dynlib_contents.append(dynlib_entry)
        dynlib_contents.append("")
        sc_lines.extend(dynlib_contents)
    if len(sc_lines) == 1:
        # Environment layer doesn't actually need customizing
        return None
    sc_contents = "\n".join(sc_lines)
    return sc_contents


def _run_postinstall(layer_path: Path) -> None:
    """Run the required post-installation steps in a deployed environment."""
    # Read the layer config file
    config = load_layer_config(layer_path)

    base_python_path = config["base_python_path"]
    env_python_path = config["python_path"]
    if base_python_path != env_python_path:
        # Generate `pyvenv.cfg` for layered environments
        # (path to base python must be emitted as an absolute path)
        venv_config = generate_pyvenv_cfg(base_python_path, config["py_version"])
        venv_config_path = layer_path / "pyvenv.cfg"
        venv_config_path.write_text(venv_config, encoding="utf-8")

        # Generate `sitecustomize.py` for layered environments
        # (avoids having to resolve relative paths on every startup)
        dynlib_paths = config["dynlib_paths"]
        sc_contents = generate_sitecustomize(config["pylib_paths"], dynlib_paths)
        if sc_contents is not None:
            sc_path = config["site_path"] / "sitecustomize.py"
            sc_path.write_text(sc_contents, encoding="utf-8")

    # Precompile Python library modules
    pylib_path = (
        layer_path / "lib"
    )  # "Lib" on Windows, but Windows is not case sensitive
    compile_dir(pylib_path, optimize=0, quiet=True)


if __name__ == "__main__":
    # Actually executing the post-installation step in a deployed environment
    _run_postinstall(Path(__file__).parent)



"""Test building the local wheels project produces the expected results."""

import os
import shutil
import subprocess
import sys
import tempfile
import venv

from pathlib import Path
from typing import Any, ClassVar


# Use unittest for these test implementations due to pytest's diff handling not working
# well for these tests, as discussed in https://github.com/pytest-dev/pytest/issues/6682
from unittest import mock

import pytest

from support import (
    DeploymentTestCase,
    EnvSummary,
    LayeredEnvSummary,
    ApplicationEnvSummary,
    SpecLoadingTestCase,
    get_os_environ_settings,
)

from venvstacks.stacks import (
    BuildEnvironment,
    PackageIndexConfig,
    StackSpec,
)
from venvstacks._util import get_env_python, run_python_command, WINDOWS_BUILD


##################################
# Sample project test helpers
##################################

_THIS_PATH = Path(__file__)
WHEEL_PROJECT_PATH = _THIS_PATH.parent / "local_wheels_project"
WHEEL_PACKAGES_PATH = WHEEL_PROJECT_PATH / "packages"
WHEEL_BUILD_REQUIREMENTS_PATH = WHEEL_PROJECT_PATH / "build-requirements.txt"
WHEEL_PROJECT_STACK_SPEC_PATH = WHEEL_PROJECT_PATH / "venvstacks.toml"
WHEEL_PROJECT_REQUIREMENTS_PATH = WHEEL_PROJECT_PATH / "requirements"
WHEEL_PROJECT_MANIFESTS_PATH = WHEEL_PROJECT_PATH / "expected_manifests"

WHEEL_PROJECT_PATHS = (
    WHEEL_PROJECT_STACK_SPEC_PATH,
    WHEEL_PROJECT_PATH / "dynlib_import.py",
    WHEEL_PROJECT_PATH / "windows_only_dynlib_import.py",
)


class _WheelBuildEnv:
    def __init__(self, working_path: Path) -> None:
        self._working_path = working_path
        self.wheel_path = wheel_path = working_path / "wheels"
        wheel_path.mkdir()
        self._venv_path = venv_path = working_path / "build_venv"
        venv.create(venv_path, symlinks=(not WINDOWS_BUILD), with_pip=True)
        self._python_path = python_path = get_env_python(venv_path)
        self._run_pip_install(
            ["-r", str(WHEEL_BUILD_REQUIREMENTS_PATH)], with_index=True
        )
        self._venv_bin_path = python_path.parent

    def remove_venv(self):
        # Test suite is done with the build, only keep the built wheels around
        shutil.rmtree(self._venv_path)

    def _run_pip(
        self, cmd_args: list[str], **kwds: Any
    ) -> subprocess.CompletedProcess[str]:
        command = [
            str(self._python_path),
            "-X",
            "utf8",
            "-Im",
            "pip",
            "--no-input",
            *cmd_args,
        ]
        return run_python_command(command, **kwds)

    def _run_pip_install(
        self, cmd_args: list[str], *, with_index: bool, **kwds: Any
    ) -> subprocess.CompletedProcess[str]:
        return self._run_pip(
            [
                "install",
                *(() if with_index else ("--no-index",)),
                "--no-deps",
                "--only-binary",
                ":all:",
                "--find-links",
                str(self.wheel_path),
                *cmd_args,
            ]
        )

    def build_wheel(self, src_path: Path) -> subprocess.CompletedProcess[str]:
        path_envvar = os.getenv("PATH", "")
        venv_bin_dir = str(self._venv_bin_path)
        env_settings = {
            "PKGCONF_PYPI_EMBEDDED_ONLY": "1",
        }
        if venv_bin_dir not in path_envvar:
            env_settings["PATH"] = f"{venv_bin_dir}{os.pathsep}{path_envvar}"
        result = self._run_pip(
            [
                "wheel",
                "--no-index",
                "--no-build-isolation",  # Build env is managed by the test suite
                "--wheel-dir",
                str(self.wheel_path),
                str(src_path),
            ],
            env=env_settings,
        )
        # Work around https://github.com/mesonbuild/meson-python/issues/639
        # (even when built on later versions, dynlibs are compatible with 3.11)
        normalized_name = src_path.name.replace("-", "_")
        expected_wheel = f"venvstacks_testing_{normalized_name}-*.whl"
        for built_wheel in self.wheel_path.glob(expected_wheel):
            built_name = built_wheel.name
            py_major, py_minor, *_ = sys.version_info
            cp_tag = f"cp{py_major}{py_minor}"
            fixed_name = built_name.replace(cp_tag, "cp311")
            if fixed_name != built_name:
                fixed_path = built_wheel.with_name(fixed_name)
                os.rename(built_wheel, fixed_path)
                print(f"Renamed {built_wheel} -> {fixed_path}")
            break
        else:
            raise RuntimeError(f"Failed to build expected wheel {expected_wheel!r}")
        return result

    def install_built_wheel(self, name: str) -> subprocess.CompletedProcess[str]:
        return self._run_pip_install([name], with_index=False)


def _build_local_wheels(working_path: Path) -> Path:
    build_env = _WheelBuildEnv(working_path)
    build_env.build_wheel(WHEEL_PACKAGES_PATH / "dynlib-publisher")
    build_env.install_built_wheel("venvstacks-testing-dynlib-publisher")
    build_env.build_wheel(WHEEL_PACKAGES_PATH / "dynlib-consumer")
    # Ensure the shared library in the build folder isn't found by the dynamic linker
    build_env.remove_venv()
    return build_env.wheel_path


def _define_build_env(
    working_path: Path, index_config: PackageIndexConfig
) -> BuildEnvironment:
    """Define a build environment for the sample project in a temporary folder."""
    # To avoid side effects from lock file creation, copy input files to the working path
    for src_path in WHEEL_PROJECT_PATHS:
        dest_path = working_path / src_path.name
        shutil.copyfile(src_path, dest_path)
    # Include "/../" in the spec path in order to test relative path resolution when
    # accessing the Python executables (that can be temperamental, especially on macOS).
    # The subdirectory won't be used for anything, so it being missing shouldn't matter.
    working_spec_path = working_path / "_unused_dir/../venvstacks.toml"
    stack_spec = StackSpec.load(working_spec_path)
    build_path = working_path / "_build🐸"
    return stack_spec.define_build_environment(build_path, index_config)


##################################
# Expected layer definitions
##################################

EXPECTED_RUNTIMES = [
    EnvSummary("cpython-3.11", ""),
]

EXPECTED_FRAMEWORKS = [
    LayeredEnvSummary("both-wheels", "framework-", "cpython-3.11", ()),
    LayeredEnvSummary("only-publisher", "framework-", "cpython-3.11", ()),
    LayeredEnvSummary(
        "only-consumer", "framework-", "cpython-3.11", ("only-publisher",)
    ),
    LayeredEnvSummary("broken-publisher", "framework-", "cpython-3.11", ()),
    LayeredEnvSummary(
        "broken-consumer", "framework-", "cpython-3.11", ("broken-publisher",)
    ),
]

EXPECTED_APPLICATIONS = [
    ApplicationEnvSummary(
        "via-combined-layer", "app-", "cpython-3.11", ("both-wheels",)
    ),
    ApplicationEnvSummary(
        "via-split-layers",
        "app-",
        "cpython-3.11",
        (
            "only-consumer",
            "only-publisher",
        ),
    ),
    ApplicationEnvSummary(
        "via-add-dll-directory",
        "app-",
        "cpython-3.11",
        (
            "broken-consumer",
            "broken-publisher",
        ),
    ),
]

EXPECTED_ENVIRONMENTS = EXPECTED_RUNTIMES.copy()
EXPECTED_ENVIRONMENTS.extend(EXPECTED_FRAMEWORKS)
EXPECTED_ENVIRONMENTS.extend(EXPECTED_APPLICATIONS)

##########################
# Test cases
##########################


class TestStackSpec(SpecLoadingTestCase):
    # Test cases that only need the stack specification file

    def test_spec_loading(self) -> None:
        self.check_stack_specification(
            WHEEL_PROJECT_STACK_SPEC_PATH,
            EXPECTED_ENVIRONMENTS,
            EXPECTED_RUNTIMES,
            EXPECTED_FRAMEWORKS,
            EXPECTED_APPLICATIONS,
        )


@pytest.mark.slow
class TestBuildEnvironment(DeploymentTestCase):
    # Test cases that need the full build environment to exist
    EXPECTED_APP_OUTPUT = "Environment launch module executed successfully"

    _wheel_temp_dir: ClassVar[tempfile.TemporaryDirectory | None] = None
    local_wheel_path: ClassVar[Path]
    working_path: Path
    build_env: BuildEnvironment

    @classmethod
    def setUpClass(cls) -> None:
        work_dir = tempfile.TemporaryDirectory()
        cls._wheel_temp_dir = work_dir
        cls.local_wheel_path = _build_local_wheels(Path(work_dir.name))

    @classmethod
    def tearDownClass(cls) -> None:
        wheel_temp_dir = cls._wheel_temp_dir
        if wheel_temp_dir is not None:
            wheel_temp_dir.cleanup()

    def setUp(self) -> None:
        working_dir = tempfile.TemporaryDirectory()
        self.addCleanup(working_dir.cleanup)
        working_path = Path(working_dir.name)
        self.working_path = working_path
        index_config = PackageIndexConfig(
            query_default_index=False, local_wheel_dirs=[self.local_wheel_path]
        )
        self.build_env = _define_build_env(working_path, index_config)
        os_env_updates = dict(get_os_environ_settings())
        # Loading local wheels, so ignore the date based lock resolution pin,
        # but allow for other env vars to be overridden
        os_env_updates.pop("UV_EXCLUDE_NEWER", None)
        os_env_patch = mock.patch.dict("os.environ", os_env_updates)
        os_env_patch.start()
        self.addCleanup(os_env_patch.stop)

    def test_create_environments(self) -> None:
        # Faster test to check the links between build envs are set up correctly
        # (if this fails, there's no point even trying the full slow test case)
        build_env = self.build_env
        build_env.create_environments()
        self.check_build_environments(self.build_env.all_environments())

    def test_locking_and_publishing(self) -> None:
        # Need long diffs to get useful output from metadata checks
        self.maxDiff = None
        # This is organised as subtests in a monolithic test sequence to reduce CI overhead
        # Separating the tests wouldn't really make them independent, unless the outputs of
        # the initial intermediate steps were checked in for use when testing the later steps.
        # Actually configuring and building the environments is executed outside the subtest
        # declarations, since actual build failures need to fail the entire test method.
        subtests_started = subtests_passed = 0  # Track subtest failures
        build_env = self.build_env
        # Create and link the layer build environments
        build_env.create_environments(lock=True)
        # Don't even try to continue if the environments aren't properly linked
        self.check_build_environments(self.build_env.all_environments())
        # Test stage: ensure exported environments allow launch module execution
        subtests_started += 1
        with self.subTest("Check environment export"):
            export_path = self.working_path / "_export🦎"
            export_result = build_env.export_environments(export_path)
            self.check_environment_exports(export_path, export_result)
            subtests_passed += 1

        # Work around pytest-subtests not failing the test case when subtests fail
        # https://github.com/pytest-dev/pytest-subtests/issues/76
        self.assertEqual(
            subtests_passed, subtests_started, "Fail due to failed subtest(s)"
        )



#!/bin/python3
"""Utility library to convert Python virtual environments to portable archives."""

# This is conceptually inspired by conda-pack (but structured somewhat differently).
# venv-pack and venv-pack2 were considered, and may still be an option in the future,
# but for now, something narrowly focused on the needs of the venvstacks project
# is the preferred option.
#
# This is primarily about reducing the number of potential sources of bugs - while
# conda-pack appears to be reasonably well used, venv-pack/venv-pack2 are much less
# popular, as there's a competing approach for regular virtual environments in
# https://github.com/cloudify-cosmo/wagon (where an archive of pre-built wheels is
# shipped to target systems, and then `venv` and `pip` in `--no-index` mode are run
# directly on the target to create the deployed virtual environments).

# Requirements:
#
# * must work on Linux, Windows, macOS
# * internal symlinks are permitted but not required (e.g. on Windows)
# * relative external symlinks to adjacent folders are similarly permitted
# * external symlinks beyond that boundary are converted to hard links
# * zip archives are used on Windows, tar.xz archives on other platforms

# Allowances/limitations:
#
# * archives for a given target platform are built on the same platform
# * all entry point scripts are removed, as Python is explicitly invoked on target systems
# * environment activation scripts are dropped from the archives rather than fixed on target
# * all RECORD files are removed, as they may reference files with build dependent hashes
#   (specifically, scripts that have their shebang lines rewritten at install time)
# * all __pycache__ folders are omitted (as their contents incorporate absolute paths)
#
# Note: stacks.py covers dropping the activation scripts and files with shebang lines
#

import os
import shutil
import sys
import tempfile
import time

from datetime import datetime, timedelta, timezone, tzinfo
from enum import StrEnum
from pathlib import Path
from typing import Any, Callable, cast, Literal, Self, TextIO

from ._injected import postinstall as _default_postinstall
from ._util import as_normalized_path, StrPath, WINDOWS_BUILD as _WINDOWS_BUILD

SymlinkInfo = tuple[Path, Path]


def convert_symlinks(
    env_dir: StrPath,
    containing_dir: StrPath | None = None,
) -> tuple[list[SymlinkInfo], list[SymlinkInfo]]:
    """Make env portable by making internal symlinks relative and external links hard.

    If set, containing path must be a parent directory of the environment path and is
    used as the boundary for creating relative symlinks instead of hard links. If not set,
    the environment path itself is used as the boundary for creating relative symlinks.

    Returns a 2-tuple containing lists of internal relative link conversions and
    external hard link conversions. Each list contains source/target Path pairs.
    """
    env_path = as_normalized_path(env_dir)
    if containing_dir is None:
        containing_path = env_path
    else:
        containing_path = as_normalized_path(containing_dir)
        if not env_path.is_relative_to(containing_path):
            raise ValueError(
                f"{str(env_path)!r} is not within {str(containing_path)!r}"
            )

    relative_links = []
    external_links = []
    # Ensure internal symlinks are relative, collect external links for hard link conversion.
    # The external links are *not* eagerly converted, so only the final link in any internal
    # symlink chains gets converted to a hard link.
    for file_path in env_path.rglob("*"):
        if not file_path.is_symlink():
            continue
        target_path = file_path.readlink()
        absolute_target_path = file_path.parent / target_path
        if not absolute_target_path.is_relative_to(containing_path):
            # Link target is outside the environment being packed,
            # so replace it with a hard link to the actual underlying file
            resolved_target_path = file_path.resolve()
            external_links.append((file_path, resolved_target_path))
            continue
        # Ensure symlinks within the containing path are relative
        expected_path = Path(
            os.path.relpath(str(absolute_target_path), start=str(file_path.parent))
        )
        if target_path == expected_path:
            # Symlink is already relative as desired
            continue
        # Convert absolute symlink to relative symlink
        file_path.unlink()
        file_path.symlink_to(expected_path)
        relative_links.append((file_path, file_path.readlink()))
    # Convert any external symlinks to a hard link instead
    for file_path, resolved_target_path in external_links:
        file_path.unlink()
        file_path.hardlink_to(resolved_target_path)
    return relative_links, external_links


def get_archive_path(archive_base_name: StrPath) -> Path:
    """Report the name of the archive that will be created for the given base name."""
    extension = ".zip" if _WINDOWS_BUILD else ".tar.xz"
    return Path(os.fspath(archive_base_name) + extension)


def _inject_postinstall_script(
    env_path: Path,
    script_name: str = "postinstall.py",
    script_source: StrPath | None = None,
) -> Path:
    venv_config_path = env_path / "pyvenv.cfg"
    if venv_config_path.exists():
        # The venv config contains absolute paths referencing the base runtime environment
        # Remove it here, let the post-install script recreate it
        venv_config_path.unlink()
    if script_source is None:
        # Nothing specified, inject the default postinstall script
        script_source = _default_postinstall.__file__
    script_path = env_path / script_name
    shutil.copy2(script_source, script_path)
    return script_path


def _supports_symlinks(target_path: Path) -> bool:
    with tempfile.TemporaryDirectory(dir=target_path) as link_check_dir:
        link_check_path = Path(link_check_dir)
        link_path = link_check_path / "dest"
        try:
            os.symlink("src", link_path)
        except OSError:
            # Failed to create symlink under the target path
            return False
    # Successfully created a symlink under the target path
    return True


def export_venv(
    source_dir: StrPath,
    target_dir: StrPath,
    run_postinstall: Callable[[Path, Path], None] | None = None,
) -> Path:
    """Export the given build environment, skipping archive creation and unpacking.

    * injects a suitable ``postinstall.py`` script for the environment being exported
    * excludes ``__pycache__`` folders (for consistency with archive publication)
    * excludes package metadata RECORD files (for consistency with archive publication)
    * excludes ``sitecustomize.py`` files (generated by the post-installation script)
    * replaces symlinks with copies on Windows or if the target doesn't support symlinks

    If supplied, *run_postinstall* is called with the path to the environment's Python
    interpreter and its postinstall script, allowing execution of the post-install
    script by the calling application. The post-install script is NOT implicitly
    executed by the export process.

    Returns the path to the exported environment.
    """
    source_path = as_normalized_path(source_dir)
    target_path = as_normalized_path(target_dir)
    excluded = shutil.ignore_patterns("__pycache__", "RECORD", "sitecustomize.py")
    # Avoid symlinks on Windows, as they need elevated privileges to create
    # Also avoid them if the target folder doesn't support symlink creation
    # (that way exports to FAT/FAT32/VFAT file systems should work, even if
    # it means some files end up getting duplicated on the target)
    # Otherwise, assume symlinks have already been converted with convert_symlinks
    target_path.mkdir(parents=True, exist_ok=True)
    publish_symlinks = not _WINDOWS_BUILD and _supports_symlinks(target_path)
    shutil.copytree(
        source_path,
        target_path,
        ignore=excluded,
        symlinks=publish_symlinks,
        dirs_exist_ok=True,
    )
    postinstall_path = _inject_postinstall_script(target_path)
    if run_postinstall is not None:
        run_postinstall(target_path, postinstall_path)
    return target_path


if _WINDOWS_BUILD:
    # No tar unpacking by default on windows, so use zipfile instead
    _DEFAULT_ARCHIVE_FORMAT = "zip"
else:
    # Everywhere else, create XZ compressed tar archives
    _DEFAULT_ARCHIVE_FORMAT = "xz"

_COMPRESSION_FORMATS = {
    "tar": "",
    "tar.bz2": "bzip2",
    "tar.gz": "gzip",
    "tar.xz": "xz",
}

ProgressCallback = Callable[[str], None]


class CompressionFormat(StrEnum):
    """Compression format for published environment."""

    UNCOMPRESSED = ""
    BZIP2 = "bzip2"
    GZIP = "gzip"
    XZ = "xz"
    ZIP = "zip"

    @classmethod
    def get_format(cls, format: str | None) -> Self:
        """Get compression format for given value."""
        if format is None:
            return cls(_DEFAULT_ARCHIVE_FORMAT)
        return cls(_COMPRESSION_FORMATS.get(format, format))

    @property
    def is_tar_format(self) -> bool:
        """Whether this compression format is for a tar archive."""
        return self is not self.ZIP

    def make_archive(
        self,
        base_name: StrPath,
        root_dir: StrPath,
        base_dir: StrPath,
        max_mtime: float | None = None,
        progress_callback: ProgressCallback | None = None,
    ) -> str:
        """Create layer archive using this archive format."""
        if self.is_tar_format:
            return _make_tar_archive(
                base_name,
                root_dir,
                base_dir,
                max_mtime,
                progress_callback,
                compress=str(self),
            )
        # Not a tar compression format -> emit a zipfile instead
        return _make_zipfile(
            base_name, root_dir, base_dir, max_mtime, progress_callback
        )


def create_archive(
    source_dir: StrPath,
    archive_base_name: StrPath,
    *,
    install_target: str | None = None,
    clamp_mtime: datetime | None = None,
    work_dir: StrPath | None = None,
    show_progress: bool = True,
    format: CompressionFormat | None = None,
) -> Path:
    """shutil.make_archive replacement, tailored for Python virtual environments.

    * injects a suitable ``postinstall.py`` script for the environment being archived
    * always creates zipfile archives on Windows and xztar archives elsewhere
    * excludes ``__pycache__`` folders (to reduce archive size and improve reproducibility)
    * excludes package metadata RECORD files (to improve reproducibility)
    * excludes ``sitecustomize.py`` files (generated by the post-installation script)
    * replaces symlinks with copies on Windows and allows external symlinks elsewhere
    * discards tar entry owner and group information
    * clears tar entry high mode bits (setuid, setgid, sticky)
    * clears tar entry group/other write mode bits
    * clamps mtime of archived files to the given clamp mtime at the latest
    * shows progress reporting by default (archiving built ML/AI libs is *slooooow*)

    Set *work_dir* if ``/tmp`` is too small for archiving tasks
    """
    archive_path = as_normalized_path(archive_base_name)
    source_path = Path(source_dir)
    if install_target is None:
        install_target = source_path.name
    with tempfile.TemporaryDirectory(dir=work_dir) as tmp_dir:
        target_path = Path(tmp_dir) / install_target
        env_path = export_venv(source_path, target_path)
        if not show_progress:

            def report_progress(_: Any) -> None:
                pass
        else:
            progress_bar = _ProgressBar()
            progress_bar.show(0.0)
            num_archive_entries = 0
            total_entries_to_archive = sum(1 for __ in env_path.rglob("*"))

            def report_progress(_: Any) -> None:
                nonlocal num_archive_entries
                num_archive_entries += 1
                progress_bar.show(num_archive_entries / total_entries_to_archive)

        max_mtime: int | None = None
        if clamp_mtime is not None:
            # We force UTC here as all builds should be happening on a filesystem that uses
            # UTC timestamps (i.e. no FAT/FAT32/VFAT allowed).
            # That means NTFS on Windows and any vaguely modern POSIX filesystem elsewhere.
            # To avoid filesystem time resolution quirks without relying on the resolution
            # details of the various archive formats, truncate mtime to exact seconds
            max_mtime = int(clamp_mtime.astimezone(timezone.utc).timestamp())
        if format is None:
            format = CompressionFormat.get_format(None)
        archive_with_extension = format.make_archive(
            archive_path, env_path.parent, env_path.name, max_mtime, report_progress
        )
        if show_progress:
            # Ensure progress bar completion is reported, even if there's a discrepancy
            # between the number of paths found by `rglob` and the number of archive entries
            progress_bar.show(1.0)
    # The name query and the archive creation should always report the same archive name
    assert archive_with_extension == os.fspath(get_archive_path(archive_base_name))
    return Path(archive_with_extension)


# Would prefer to use shutil.make_archive, but the way it works doesn't quite fit this case
# _make_tar_archive below is adjusted to be similar to make_archive, but adapted from
# https://github.com/python/cpython/blob/99d945c0c006e3246ac00338e37c443c6e08fc5c/Lib/shutil.py#L930
# to work around the limitations mentioned in https://github.com/python/cpython/issues/120036
# Puts this utility module under the Python License, but the runtime layers already include
# CPython, so also using it in the build utility doesn't introduce any new licensing concerns


def _make_tar_archive(
    base_name: StrPath,
    root_dir: StrPath,
    base_dir: StrPath,
    max_mtime: float | None = None,
    progress_callback: ProgressCallback | None = None,
    *,
    compress: str = "xz",
) -> str:
    """Create a (possibly compressed) tar file from all the files under 'base_dir'.

    'compress' must be "gzip", "bzip2", "xz", or None.

    Owner and group info is always set to 0/"root" as per
    https://reproducible-builds.org/docs/archives/.

    The output tar file will be named 'base_name' +  ".tar", possibly plus
    the appropriate compression extension (".gz", ".bz2", or ".xz").

    Returns the output filename.
    """
    import tarfile  # lazy import since ideally shutil would handle everything

    # Type checkers complain if the tar_mode string is built dynamically
    tar_mode: Literal["w", "w:gz", "w:bz2", "w:xz"]
    if compress is None or compress == "":
        tar_mode = "w"
        compress_ext = ""
    elif compress == "gzip":
        tar_mode = "w:gz"
        compress_ext = ".gz"
    elif compress == "bzip2":
        tar_mode = "w:bz2"
        compress_ext = ".bz2"
    elif compress == "xz":
        tar_mode = "w:xz"
        compress_ext = ".xz"
    else:
        raise ValueError(
            "bad value for 'compress', or compression format not "
            "supported : {0}".format(compress)
        )

    archive_name = os.fspath(base_name) + ".tar" + compress_ext
    archive_dir = os.path.dirname(archive_name)

    if archive_dir and not os.path.exists(archive_dir):
        os.makedirs(archive_dir)

    # Clamp mtime only if requested
    # Force to an int to keep mypy happy: https://github.com/python/typeshed/issues/12520
    # Once a mypy update is published with that issue fixed, this workaround can
    # be replaced by a minimum mypy version requirement in the dev dependencies.
    if max_mtime is None:
        _clamp_mtime = None
    else:
        truncated_max_mtime = int(max_mtime)

        def _clamp_mtime_impl(mtime: int | float) -> int:
            # pyright has a newer typeshed than mypy, so resort to a cast
            # on the input value until this entire workaround can be dropped
            return min(truncated_max_mtime, cast(int, mtime))

        # Work around for https://github.com/microsoft/pyright/issues/9114
        _clamp_mtime = _clamp_mtime_impl

    # Ensure archive entries are reproducible across repeated builds
    def _process_archive_entry(tarinfo: tarfile.TarInfo) -> tarfile.TarInfo:
        # Omit owner & group info from build system
        tarinfo.uid = tarinfo.gid = 0
        tarinfo.uname = tarinfo.gname = "root"
        if _clamp_mtime is not None:
            tarinfo.mtime = _clamp_mtime(tarinfo.mtime)
        # Ensure permissions are compatible with `tar_filter` extraction
        # Layered environments will still need to be fully trusted when
        # unpacking them (due to the external symlinks to the base runtime)
        mode = tarinfo.mode
        if mode is not None:
            # Apply the same mode filtering as tarfile.tar_filter in 3.12+
            # https://docs.python.org/3.13/library/tarfile.html#tarfile.tar_filter
            # Clears high bits (e.g. setuid/setgid), and the group/other write bits
            tarinfo.mode = mode & 0o755
        # Report progress if requested
        if progress_callback is not None:
            progress_callback(tarinfo.name)
        return tarinfo

    # creating the tarball
    tar = tarfile.open(archive_name, tar_mode)
    arcname = base_dir
    if root_dir is not None:
        base_dir = os.path.join(root_dir, base_dir)
    try:
        # In Python 3.7+, tar.add inherently adds entries in sorted order
        tar.add(base_dir, arcname, filter=_process_archive_entry)
    finally:
        tar.close()

    if root_dir is not None:
        archive_name = os.path.abspath(archive_name)
    return archive_name


# _make_zipfile below is adjusted to be similar to make_archive, but adapted from
# https://github.com/python/cpython/blob/99d945c0c006e3246ac00338e37c443c6e08fc5c/Lib/shutil.py#L1000

if _WINDOWS_BUILD:

    def _set_mtime(fspath: str, mtime: int | float) -> None:
        # There's no `follow_symlinks` option available on Windows
        os.utime(fspath, (mtime, mtime))
else:

    def _set_mtime(fspath: str, mtime: int | float) -> None:
        os.utime(fspath, (mtime, mtime), follow_symlinks=False)


def _make_zipfile(
    base_name: StrPath,
    root_dir: StrPath,
    base_dir: StrPath,
    max_mtime: float | None = None,
    progress_callback: ProgressCallback | None = None,
) -> str:
    """Create a zip file from all the files under 'base_dir'.

    The output zip file will be named 'base_name' + ".zip".  Returns the
    name of the output zip file.
    """
    import zipfile  # lazy import since ideally shutil would handle everything

    zip_filename = os.fspath(base_name) + ".zip"
    archive_dir = os.path.dirname(base_name)

    if archive_dir and not os.path.exists(archive_dir):
        os.makedirs(archive_dir)

    # Unlike _make_tar_archive, progress is reported from multiple places,
    # so define a dummy callback if no actual callback is given
    if progress_callback is None:

        def _default_progress_callback(_: Any) -> None:
            pass

        # Work around for https://github.com/microsoft/pyright/issues/9114
        progress_callback = _default_progress_callback

    # zipfile stores local timestamps: https://github.com/python/cpython/issues/123059
    # We don't want that, but zipfile doesn't currently provide a nice API to adjust the
    # timestamps when adding files to the archive, so we instead intentionally make the
    # filesystem timestamps *wrong* such that calling `time.localtime` reports a UTC time
    need_mtime_adjustment = time.localtime().tm_gmtoff != 0
    if not need_mtime_adjustment:
        # Local time is UTC anyway, so no timezone adjustment is needed
        def adjust_mtime(mtime: float) -> float:
            return mtime
    else:
        # Adjust filesystem mtime so `zipfile` sets the desired value in the archive entry
        # casts are needed due to https://github.com/python/mypy/issues/10067
        local_tz = cast(tzinfo, datetime.now().astimezone().tzinfo)
        local_tz_offset = cast(timedelta, local_tz.utcoffset(None))

        def adjust_mtime(mtime: float) -> float:
            # mtime is given here in UTC time. To get `zipfile` to see that time value
            # when calling `time.localtime`, we need to do a local -> UTC conversion on the
            # UTC timestamp, so `zipfile`'s UTC -> local conversion gives back the UTC time
            local_mtime = datetime.fromtimestamp(mtime).astimezone()
            adjusted_mtime = local_mtime - local_tz_offset
            return adjusted_mtime.timestamp()

    with zipfile.ZipFile(zip_filename, "w", compression=zipfile.ZIP_DEFLATED) as zf:
        if max_mtime is None:
            # Clamp mtime only if requested
            max_mtime = float("inf")

        def _add_zip_entry(fspath: str, arcname: str) -> None:
            fs_mtime = os.lstat(fspath).st_mtime
            zip_entry_mtime = adjust_mtime(min(fs_mtime, max_mtime))
            if zip_entry_mtime != fs_mtime:
                _set_mtime(fspath, zip_entry_mtime)
            zf.write(fspath, arcname)

        arcname = os.path.normpath(base_dir)
        if root_dir is not None:
            base_dir = os.path.join(root_dir, base_dir)
        base_dir = os.path.normpath(base_dir)
        if arcname != os.curdir:
            _add_zip_entry(base_dir, arcname)
        # Python 3.11 compatibility: use os.walk instead of Path.walk
        for dirpath, dirnames, filenames in os.walk(base_dir):
            arcdirpath = dirpath
            if root_dir is not None:
                arcdirpath = os.path.relpath(arcdirpath, root_dir)
            arcdirpath = os.path.normpath(arcdirpath)
            dirnames.sort()  # Ensure recursion occurs in a consistent order
            for name in dirnames:
                path = os.path.join(dirpath, name)
                arcname = os.path.join(arcdirpath, name)
                _add_zip_entry(path, arcname)
                progress_callback(name)
            for name in sorted(filenames):
                path = os.path.join(dirpath, name)
                path = os.path.normpath(path)
                if os.path.isfile(path):
                    arcname = os.path.join(arcdirpath, name)
                    _add_zip_entry(path, arcname)
                    progress_callback(name)

    if root_dir is not None:
        zip_filename = os.path.abspath(zip_filename)
    return zip_filename


# Basic progress bar support, taken from ncoghlan's SO answer at
# https://stackoverflow.com/questions/3160699/python-progress-bar/78590319#78590319
# (since the code originated with her, and she also implemented it here,
# it isn't subject to Stack Overflow's CC-BY-SA terms)
#
# Archiving pytorch (and similarly large AI/ML libraries) takes a long time,
# so you really need some assurance that progress is being made.
#
# If compression times are a significant problem, it would be worth moving in the same
# direction as conda-pack did, and implementing support for parallel compression (the
# compression libraries all drop the GIL when compressing data chunks, so this approach
# scales effectively up to the number of available CPUs)
#
# See https://github.com/lmstudio-ai/venvstacks/issues/4

_ProgressSummary = tuple[int, str]
_ProgressReport = tuple[str, _ProgressSummary]


class _ProgressBar:
    """Display & update a progress bar."""

    TEXT_ABORTING = "Aborting..."
    TEXT_COMPLETE = "Complete!"
    TEXT_PROGRESS = "Archiving"

    bar_length: int
    stream: TextIO
    _last_displayed_text: str | None
    _last_displayed_summary: _ProgressSummary | None

    def __init__(self, bar_length: int = 25, stream: TextIO = sys.stdout) -> None:
        self.bar_length = bar_length
        self.stream = stream
        self._last_displayed_text = None
        self._last_displayed_summary = None

    def reset(self) -> None:
        """Forget any previously displayed text (affects subsequent call to show())."""
        self._last_displayed_text = None
        self._last_displayed_summary = None

    def _format_progress(self, progress: float, aborting: bool) -> _ProgressReport:
        """Internal helper that also reports the number of completed increments."""
        bar_length = self.bar_length
        progress = float(progress)
        if progress >= 1:
            # Report task completion
            completed_increments = bar_length
            status = " " + self.TEXT_COMPLETE
            progress = 1.0
        else:
            # Truncate progress to ensure bar only fills when complete
            completed_increments = int(progress * bar_length)
            status = (" " + self.TEXT_ABORTING) if aborting else ""
        remaining_increments = bar_length - completed_increments
        bar_content = f"{'#' * completed_increments}{'-' * remaining_increments}"
        percentage = f"{progress * 100:.2f}"
        progress_text = f"{self.TEXT_PROGRESS}: [{bar_content}] {percentage}%{status}"
        return progress_text, (completed_increments, status)

    def format_progress(self, progress: float, *, aborting: bool = False) -> str:
        """Format progress bar, percentage, and status for given fractional progress."""
        return self._format_progress(progress, aborting)[0]

    def show(self, progress: float, *, aborting: bool = False) -> None:
        """Display the current progress on the console."""
        progress_text, progress_summary = self._format_progress(progress, aborting)
        if progress_text == self._last_displayed_text:
            # No change to display output, so skip writing anything
            # (this reduces overhead on both interactive and non-interactive streams)
            return
        interactive = self.stream.isatty()
        if not interactive and progress_summary == self._last_displayed_summary:
            # For non-interactive streams, skip output if only the percentage has changed
            # (this avoids flooding the output on non-interactive streams that ignore '\r')
            return
        if not interactive or aborting or progress >= 1:
            # Final output or non-interactive output, so advance to next line
            line_end = "\n"
        else:
            # Interactive progress output, so try to return to start of current line
            line_end = "\r"
        sys.stdout.write(progress_text + line_end)
        sys.stdout.flush()
        self._last_displayed_text = progress_text
        self._last_displayed_summary = progress_summary



"""Test cases for dynamic library discovery and symlinking."""

import sys
import tempfile

import pytest

from pathlib import Path
from importlib.machinery import EXTENSION_SUFFIXES
from typing import Generator, cast

from venvstacks._util import (
    find_shared_libraries,
    map_symlink_targets,
)


@pytest.fixture
def temp_dir_path() -> Generator[Path, None, None]:
    with tempfile.TemporaryDirectory() as dir_name:
        yield Path(dir_name)


@pytest.mark.skipif(
    sys.platform == "win32", reason="No scan for dynamic libraries on Windows"
)
def test_find_shared_libraries(temp_dir_path: Path) -> None:
    # Scanning for shared libraries is entirely name based
    base_names = (
        "libpng",
        "libcodec",
        "libother",
    )
    linked_extensions = (".so", ".dylib")
    all_extensions = {*EXTENSION_SUFFIXES, *linked_extensions}
    expected_names = [name + ext for name in base_names for ext in linked_extensions]
    all_names = [name + ext for name in base_names for ext in all_extensions]
    subdir_paths = (
        temp_dir_path / "project1",
        temp_dir_path / "project2/libs",
        temp_dir_path / "project3/_files/libs",
    )
    expected_paths = {p / name for p in subdir_paths for name in expected_names}
    dynlib_paths = [p / name for p in subdir_paths for name in all_names]
    for p in dynlib_paths:
        p.parent.mkdir(parents=True, exist_ok=True)
        p.touch()

    py_version_info = cast(tuple[str, str], tuple(map(str, sys.version_info[:2])))
    paths_without_exclusions = {*find_shared_libraries(py_version_info, temp_dir_path)}
    assert paths_without_exclusions == expected_paths

    exclusions = [
        "project1/**/libpng.*",
        "project2/**/libcodec.*",
        "project3/**/libother.*",
    ]
    paths_with_exclusions = {
        *find_shared_libraries(py_version_info, temp_dir_path, excluded=exclusions)
    }

    def _exclude_path(p: Path) -> bool:
        return (
            (p.stem == "libpng" and "project1" in p.parts)
            or (p.stem == "libcodec" and "project2" in p.parts)
            or (p.stem == "libother" and "project3" in p.parts)
        )

    expected_paths_with_exclusions = {p for p in expected_paths if not _exclude_path(p)}
    assert paths_with_exclusions == expected_paths_with_exclusions


def test_map_symlink_targets() -> None:
    # Symlink target mapping doesn't check if the paths exist
    # (that's handled when creating the list of inputs)
    symlink_dir_path = Path()
    # No input -> no output
    assert map_symlink_targets(symlink_dir_path, []) == ({}, {})

    # No ambiguity
    targets_dir_path = symlink_dir_path / "targets"
    target_paths = [targets_dir_path / f"target{n}" for n in range(5)]
    symlink_paths = [symlink_dir_path / p.name for p in target_paths]
    expected_mapping = dict(zip(symlink_paths, target_paths))
    assert map_symlink_targets(symlink_dir_path, target_paths) == (expected_mapping, {})

    # Ambiguity
    conflicting_path = targets_dir_path / "conflict" / target_paths[0].name
    ambiguous_paths = [*target_paths, *target_paths, conflicting_path]
    expected_valid_mapping = dict(zip(symlink_paths[1:], target_paths[1:]))
    expected_ambiguity = {symlink_paths[0]: {target_paths[0], conflicting_path}}
    assert map_symlink_targets(symlink_dir_path, ambiguous_paths) == (
        expected_valid_mapping,
        expected_ambiguity,
    )



"""Command line interface implementation."""

import os.path
import sys

from typing import Annotated, Iterable, Sequence

import typer

from .stacks import (
    StackSpec,
    BuildEnvironment,
    _format_json,
    PackageIndexConfig,
    EnvNameBuild,
)

# Inspired by the Python 3.13+ `argparse` feature,
# but reports `python -m venvstacks` whenever `__main__`
# refers to something other than the entry point script,
# rather than trying to infer anything from the main
# module's `__spec__` attributes.
_THIS_PACKAGE = __spec__.parent


def _get_usage_name() -> str:
    exec_name = os.path.basename(sys.argv[0]).removesuffix(".exe")
    if exec_name == _THIS_PACKAGE:
        # Entry point wrapper, suggest direct execution
        return exec_name
    # Could be `python -m`, could be the test suite,
    # could be something else calling `venvstacks.cli.main`,
    # but treat it as `python -m venvstacks` regardless
    py_name = os.path.basename(sys.executable).removesuffix(".exe")
    return f"{py_name} -m {_THIS_PACKAGE}"


_cli = typer.Typer(
    name=_get_usage_name(),
    add_completion=False,
    pretty_exceptions_show_locals=False,
    no_args_is_help=True,
)


@_cli.callback()
def handle_app_options() -> None:
    """Lock, build, and publish Python virtual environment stacks."""
    # TODO: Handle app logging config via main command level options
    #       Part of https://github.com/lmstudio-ai/venvstacks/issues/5
    pass


# The shared CLI argument and option annotations have to be module level globals,
# otherwise pylance complains about runtime variable use in type annotations
#
# Defined prefixes:
#
#   * _CLI_ARG: required argument (no option name, arbitrary type accepting a string arg)
#   * _CLI_OPT_FLAG: boolean option, must be True or False
#   * _CLI_OPT_TRISTATE: boolean option, but allows None to indicate "not set"
#   * _CLI_OPT_STR: optional string (defaulting to empty string to indicate "not set")
#   * _CLI_OPT_STRLIST: multi-value list of strings
#
# The unit tests ensure the internal consistency of the CLI command annotations

# Format skips below keep arg annotations from being collapsed onto fewer lines

# Required arguments: where to find the stack spec
_CLI_ARG_spec_path = Annotated[
    str,
    typer.Argument(help="Path to TOML file specifying layers to build")
]  # fmt: skip

# Optional directory arguments: where to put build and output artifacts
_CLI_OPT_STR_build_dir = Annotated[
    str,
    typer.Option(help="Directory (relative to spec) for intermediate build artifacts"),
]  # fmt: skip
_CLI_OPT_STR_output_dir = Annotated[
    str,
    typer.Option(help="Directory (relative to spec) for output artifacts and metadata"),
]  # fmt: skip

# Build pipeline steps: cleaning, locking, building, and publishing archives
_CLI_OPT_FLAG_clean = Annotated[
    bool,
    typer.Option(help="Remove existing environments before building")
]  # fmt: skip
_CLI_OPT_FLAG_lock = Annotated[
    bool,
    typer.Option(help="Update layer lock files before building")
]  # fmt: skip
_CLI_OPT_FLAG_build = Annotated[
    bool,
    typer.Option(help="Build layer environments")
]  # fmt: skip
_CLI_OPT_FLAG_publish = Annotated[
    bool,
    typer.Option(help="Create archives from built environments")
]  # fmt: skip

# Configuring the input artifact retrieval process
_CLI_OPT_FLAG_index = Annotated[
    bool,
    typer.Option(
        help="Query the default package index (PyPI) for installation artifacts"
    ),
]  # fmt: skip
_CLI_OPT_STRLIST_local_wheels = Annotated[
    list[str] | None,
    typer.Option(
        help="Additional directory (relative to spec) for locally built wheel archives"
    ),
]  # fmt: skip

# Adjust naming of published archives and metadata files
_CLI_OPT_FLAG_tag_outputs = Annotated[
    bool,
    typer.Option(help="Include platform and other details in published names")
]  # fmt: skip

# Archive publication options for the publish subcommand
_CLI_OPT_FLAG_force = Annotated[
    bool,
    typer.Option(help="Publish archive even if input metadata is unchanged")
]  # fmt: skip
_CLI_OPT_FLAG_dry_run = Annotated[
    bool,
    typer.Option(help="List archives that would be published")
]  # fmt: skip

# Selective processing of defined layers
_CLI_OPT_STRLIST_include = Annotated[
    list[str] | None,
    typer.Option(
        help="Include specified layer in requested operations.\n"
        "Option may be supplied multiple times to match multiple layer names.\n"
        "Also accepts Python 'fnmatch' syntax to match multiple layer names.\n"
        "If this option is omitted, all defined layers are included."
    ),
]
_CLI_OPT_STRLIST_reset_lock = Annotated[
    list[str] | None,
    typer.Option(
        help="Delete existing layer lock file before locking specified layer.\n"
        "Option may be supplied multiple times to match multiple layer names.\n"
        "Also accepts Python 'fnmatch' syntax to match multiple layer names.\n"
        "Only layers selected for locking will have their lock files reset.\n"
        "If this option is omitted, packages are only updated as necessary."
    ),
]
_CLI_OPT_FLAG_allow_missing = Annotated[
    bool,
    typer.Option(help="Allow layer filtering entries that do not match any layers")
]  # fmt: skip

# Handling layers that included layers depend on
_CLI_OPT_FLAG_lock_dependencies = Annotated[
    bool,
    typer.Option(help="Also lock dependencies of included layers")
]  # fmt: skip
_CLI_OPT_FLAG_build_dependencies = Annotated[
    bool,
    typer.Option(help="Also build dependencies of included layers")
]  # fmt: skip
_CLI_OPT_FLAG_publish_dependencies = Annotated[
    bool,
    typer.Option(help="Also publish dependencies of included layers")
]  # fmt: skip
_CLI_OPT_TRISTATE_include_dependencies = Annotated[
    bool | None,
    typer.Option(help="All operations include dependencies of included layers"),
]  # fmt: skip

# Handling layers that depend on included layers
# Note: when locking, layers that depend on included layers are *always* relocked
_CLI_OPT_FLAG_build_derived = Annotated[
    bool,
    typer.Option(help="Also build environments that depend on included layers")
]  # fmt: skip
_CLI_OPT_FLAG_publish_derived = Annotated[
    bool,
    typer.Option(help="Also publish archives that depend on included layers")
]  # fmt: skip


def _define_build_environment(
    spec_path: str,
    build_path: str,
    *,
    index: bool,
    local_wheels: Sequence[str] | None,
) -> BuildEnvironment:
    """Load given stack specification and define a build environment."""
    stack_spec = StackSpec.load(spec_path)
    index_config = PackageIndexConfig(
        query_default_index=index,
        local_wheel_dirs=local_wheels,
    )
    return stack_spec.define_build_environment(build_path, index_config)


def _handle_layer_include_options(
    build_env: BuildEnvironment,
    include: Sequence[str] | None,
    *,
    allow_missing: bool,
    lock: bool,
    build: bool,
    publish: bool,
    lock_dependencies: bool,
    build_dependencies: bool,
    publish_dependencies: bool,
    build_derived: bool,
    publish_derived: bool,
    reset_locks: Sequence[str] | None,
) -> None:
    included_layers: Iterable[EnvNameBuild]
    if not include:
        # Include all layers (helper invocation is for lock reset filtering)
        included_layers = [env.env_name for env in build_env.all_environments()]
    else:
        included_layers, unmatched_patterns = build_env.filter_layers(include)
        if unmatched_patterns:
            err_details = f"No matching layers found for: {unmatched_patterns!r}"
            if allow_missing:
                print(f"WARNING: {err_details}")
            else:
                warning_hint = "Pass '--allow-missing' to convert to warning"
                print(f"ERROR: {err_details}\n  {warning_hint}")
                raise typer.Exit(code=1)
    layers_to_reset: Iterable[EnvNameBuild]
    if not reset_locks:
        # No lock reset patterns -> don't reset any layers
        layers_to_reset = ()
    else:
        layers_to_reset, unmatched = build_env.filter_layers(reset_locks)
        if unmatched:
            err_details = f"No matching layers found for upgrade: {unmatched!r}"
            if allow_missing:
                print(f"WARNING: {err_details}")
            else:
                warning_hint = "Pass '--allow-missing' to convert to warning"
                print(f"ERROR: {err_details}\n  {warning_hint}")
                raise typer.Exit(code=1)
    # Layer lock files are only reset if the layer actually gets locked,
    # so there's no explicit cross-check between included and reset layers
    build_env.select_layers(
        included_layers,
        lock=lock,
        build=build,
        publish=publish,
        lock_dependencies=lock_dependencies,
        build_dependencies=build_dependencies,
        publish_dependencies=publish_dependencies,
        build_derived=build_derived,
        publish_derived=publish_derived,
        reset_locks=layers_to_reset,
    )


def _publication_dry_run(
    build_env: BuildEnvironment,
    output_dir: str,
    tag_outputs: bool,
) -> None:
    base_output_path, dry_run_result = build_env.publish_artifacts(
        output_dir,
        dry_run=True,
        tag_outputs=tag_outputs,
    )
    print("Archive creation skipped, reporting publishing request details:")
    print(_format_json(dry_run_result))
    print(f"Archives and metadata will be published to {base_output_path}")
    print("Archive creation skipped (specify --publish to request archive build)")


def _publish_artifacts(
    build_env: BuildEnvironment,
    output_dir: str,
    force: bool,
    dry_run: bool,
    tag_outputs: bool,
) -> None:
    if dry_run:
        _publication_dry_run(build_env, output_dir, tag_outputs=tag_outputs)
        return
    manifest_path, snippet_paths, archive_paths = build_env.publish_artifacts(
        output_dir,
        force=force,
        tag_outputs=tag_outputs,
    )
    base_output_path = os.path.commonpath(
        [manifest_path, *snippet_paths, *archive_paths]
    )
    print(manifest_path.read_text(encoding="utf-8"))
    print(
        f"Full stack manifest saved to: {manifest_path.relative_to(base_output_path)}"
    )
    print("Individual layer manifests:")
    for snippet_path in snippet_paths:
        print(f"  {snippet_path.relative_to(base_output_path)}")
    print("Generated artifacts:")
    for archive_path in archive_paths:
        print(f"  {archive_path.relative_to(base_output_path)}")
    print(f"All paths reported relative to {base_output_path}")


def _export_dry_run(
    build_env: BuildEnvironment,
    output_dir: str,
) -> None:
    base_output_path, dry_run_result = build_env.export_environments(
        output_dir,
        dry_run=True,
    )
    print("Environment export skipped, reporting local export request details:")
    print(_format_json(dry_run_result))
    print(f"Environments will be exported to {base_output_path}")
    print("Environment export skipped (specify --publish to request local export)")


def _export_environments(
    build_env: BuildEnvironment,
    output_dir: str,
    force: bool,
    dry_run: bool,
) -> None:
    if dry_run:
        _export_dry_run(build_env, output_dir)
        return
    manifest_path, snippet_paths, env_paths = build_env.export_environments(
        output_dir,
        force=force,
    )
    base_output_path = os.path.commonpath([manifest_path, *snippet_paths, *env_paths])
    print(manifest_path.read_text(encoding="utf-8"))
    print(
        f"Full export manifest saved to: {manifest_path.relative_to(base_output_path)}"
    )
    print("Individual manifests:")
    for snippet_path in snippet_paths:
        print(f"  {snippet_path.relative_to(base_output_path)}")
    print("Exported environments:")
    for env_path in env_paths:
        print(f"  {env_path.relative_to(base_output_path)}")
    print(f"All paths reported relative to {base_output_path}")


@_cli.command()
def build(
    # Required arguments: where to find the stack spec
    spec_path: _CLI_ARG_spec_path,
    # Optional directory arguments: where to put build and output artifacts
    build_dir: _CLI_OPT_STR_build_dir = "_build",
    output_dir: _CLI_OPT_STR_output_dir = "_artifacts",
    # Pipeline steps: cleaning, locking, building, and publishing archives
    clean: _CLI_OPT_FLAG_clean = False,
    lock: _CLI_OPT_FLAG_lock = False,
    build: _CLI_OPT_FLAG_build = True,
    publish: _CLI_OPT_FLAG_publish = False,
    # Package index access configuration
    index: _CLI_OPT_FLAG_index = True,
    local_wheels: _CLI_OPT_STRLIST_local_wheels = None,
    # Adjust naming of published archives and metadata files
    tag_outputs: _CLI_OPT_FLAG_tag_outputs = False,
    # Selective processing of defined layers
    include: _CLI_OPT_STRLIST_include = None,
    reset_lock: _CLI_OPT_STRLIST_reset_lock = None,
    allow_missing: _CLI_OPT_FLAG_allow_missing = False,
    # Handling layers that included layers depend on
    lock_dependencies: _CLI_OPT_FLAG_lock_dependencies = False,
    build_dependencies: _CLI_OPT_FLAG_build_dependencies = False,
    publish_dependencies: _CLI_OPT_FLAG_publish_dependencies = False,
    include_dependencies: _CLI_OPT_TRISTATE_include_dependencies = None,
    # Handling layers that depend on included layers
    # Note: when locking, layers that depend on included layers are *always* relocked
    build_derived: _CLI_OPT_FLAG_build_derived = False,
    publish_derived: _CLI_OPT_FLAG_publish_derived = False,
) -> None:
    """Build (/lock/publish) Python virtual environment stacks."""
    if include_dependencies is not None:
        # Override the per-operation settings
        lock_dependencies = build_dependencies = publish_dependencies = (
            include_dependencies
        )
    build_env = _define_build_environment(
        spec_path,
        build_dir,
        index=index,
        local_wheels=local_wheels,
    )
    # Update the various `want_*` flags on each environment
    # Note: CLI `publish` controls the `dry_run` flag on the `publish_artifacts` method call
    if include or reset_lock:
        _handle_layer_include_options(
            build_env,
            include,
            allow_missing=allow_missing,
            lock=lock,
            build=build,
            publish=True,
            lock_dependencies=lock_dependencies,
            build_dependencies=build_dependencies,
            publish_dependencies=publish_dependencies,
            build_derived=build_derived,
            publish_derived=publish_derived,
            reset_locks=reset_lock,
        )
    else:
        build_env.select_operations(
            lock=lock,
            build=build,
            publish=True,
        )
    build_env.create_environments(clean=clean, lock=lock)
    _publish_artifacts(
        build_env, output_dir, dry_run=not publish, force=clean, tag_outputs=tag_outputs
    )


@_cli.command()
def lock(
    # Required arguments: where to find the stack spec, and where to emit any output files
    spec_path: _CLI_ARG_spec_path,
    # Optional directory arguments: where to put build artifacts (must create envs to lock them)
    build_dir: _CLI_OPT_STR_build_dir = "_build",
    # Pipeline steps: cleaning is the only optional step for this subcommand
    clean: _CLI_OPT_FLAG_clean = False,
    # Package index access configuration
    index: _CLI_OPT_FLAG_index = True,
    local_wheels: _CLI_OPT_STRLIST_local_wheels = None,
    # Selective processing of defined layers
    include: _CLI_OPT_STRLIST_include = None,
    reset_lock: _CLI_OPT_STRLIST_reset_lock = None,
    allow_missing: _CLI_OPT_FLAG_allow_missing = False,
    # Whether to lock the layers that the included layers depend on
    lock_dependencies: _CLI_OPT_FLAG_lock_dependencies = False,
    # When locking, layers that depend on included layers are *always* relocked
) -> None:
    """Lock layer requirements for Python virtual environment stacks."""
    build_env = _define_build_environment(
        spec_path,
        build_dir,
        index=index,
        local_wheels=local_wheels,
    )
    # Update the various `want_*` flags on each environment
    if include or reset_lock:
        _handle_layer_include_options(
            build_env,
            include,
            allow_missing=allow_missing,
            lock=True,
            build=False,
            publish=False,
            lock_dependencies=lock_dependencies,
            build_dependencies=False,
            publish_dependencies=False,
            build_derived=False,
            publish_derived=False,
            reset_locks=reset_lock,
        )
    else:
        build_env.select_operations(
            lock=True,
            build=False,
            publish=False,
        )
    lock_results = build_env.lock_environments(clean=clean)
    if not lock_results:
        print("No environments found to lock")
    else:
        base_output_path = os.path.commonpath(
            [env.locked_requirements_path for env in lock_results]
        )
        print("Locked environments:")
        for env_lock in lock_results:
            relative_path = env_lock.locked_requirements_path.relative_to(
                base_output_path
            )
            print(f"  {relative_path} (locked at: {env_lock.locked_at})")
        print(f"All paths reported relative to {base_output_path}")


@_cli.command()
def publish(
    # Required arguments: where to find the stack spec, and where to emit any output files
    spec_path: _CLI_ARG_spec_path,
    # Optional directory arguments: where to find build artifacts and put output artifacts
    build_dir: _CLI_OPT_STR_build_dir = "_build",
    output_dir: _CLI_OPT_STR_output_dir = "_artifacts",
    # Optional behaviour
    force: _CLI_OPT_FLAG_force = False,
    dry_run: _CLI_OPT_FLAG_dry_run = False,
    # Adjust naming of published archives and metadata files
    tag_outputs: _CLI_OPT_FLAG_tag_outputs = False,
    # Selective processing of defined layers
    include: _CLI_OPT_STRLIST_include = None,
    allow_missing: _CLI_OPT_FLAG_allow_missing = False,
    # Handling layers that included layers depend on
    publish_dependencies: _CLI_OPT_FLAG_publish_dependencies = False,
    # Handling layers that depend on included layers
    publish_derived: _CLI_OPT_FLAG_publish_derived = False,
) -> None:
    """Publish layer archives for Python virtual environment stacks."""
    build_env = _define_build_environment(
        spec_path,
        build_dir,
        # No locking or build steps will be invoked on the environment
        index=False,
        local_wheels=None,
    )
    # Update the various `want_*` flags on each environment
    # Note: CLI `publish` controls the `dry_run` flag on the `publish_artifacts` method call
    if include:
        _handle_layer_include_options(
            build_env,
            include,
            reset_locks=None,
            allow_missing=allow_missing,
            lock=False,
            build=False,
            publish=True,
            lock_dependencies=False,
            build_dependencies=False,
            publish_dependencies=publish_dependencies,
            build_derived=False,
            publish_derived=publish_derived,
        )
    else:
        build_env.select_operations(
            lock=False,
            build=False,
            publish=True,
        )
    _publish_artifacts(
        build_env, output_dir, force=force, dry_run=dry_run, tag_outputs=tag_outputs
    )


@_cli.command()
def local_export(
    # Required arguments: where to find the stack spec, and where to emit any output files
    spec_path: _CLI_ARG_spec_path,
    # Optional directory arguments: where to find build artifacts and put output artifacts
    build_dir: _CLI_OPT_STR_build_dir = "_build",
    output_dir: _CLI_OPT_STR_output_dir = "_export",
    # Optional behaviour
    force: _CLI_OPT_FLAG_force = False,
    dry_run: _CLI_OPT_FLAG_dry_run = False,
    # Selective processing of defined layers
    include: _CLI_OPT_STRLIST_include = None,
    allow_missing: _CLI_OPT_FLAG_allow_missing = False,
    # Handling layers that included layers depend on
    publish_dependencies: _CLI_OPT_FLAG_publish_dependencies = False,
    # Handling layers that depend on included layers
    publish_derived: _CLI_OPT_FLAG_publish_derived = False,
) -> None:
    """Export layer environments for Python virtual environment stacks."""
    build_env = _define_build_environment(
        spec_path,
        build_dir,
        # No locking or build steps will be invoked on the environment
        index=False,
        local_wheels=None,
    )
    # Update the various `want_*` flags on each environment
    # Note: CLI `publish` controls the `dry_run` flag on the `publish_artifacts` method call
    if include:
        _handle_layer_include_options(
            build_env,
            include,
            reset_locks=None,
            allow_missing=allow_missing,
            lock=False,
            build=False,
            publish=True,
            lock_dependencies=False,
            build_dependencies=False,
            publish_dependencies=publish_dependencies,
            build_derived=False,
            publish_derived=publish_derived,
        )
    else:
        build_env.select_operations(
            lock=False,
            build=False,
            publish=True,
        )
    _export_environments(
        build_env,
        output_dir,
        force=force,
        dry_run=dry_run,
    )


def main(args: Sequence[str] | None = None) -> None:
    """Run the ``venvstacks`` CLI.

    If *args* is not given, defaults to using ``sys.argv``.
    """
    # Indirectly calls the relevant click.Command variant's `main` method
    # See https://click.palletsprojects.com/en/8.1.x/api/#click.BaseCommand.main
    _cli(args, windows_expand_args=False)



import sys
import dynlib_publisher  # Needed for os.add_dll_directory on Windows

EXPECT_CONSUMER_IMPORT = sys.platform == "win32"

try:
    import dynlib_consumer
except ImportError:
    if EXPECT_CONSUMER_IMPORT:
        raise
else:
    if not EXPECT_CONSUMER_IMPORT:
        sys.exit("Successful import when it wasn't expected")


if __name__ == "__main__":
    # The app environment should NOT have access to pip
    from importlib.util import find_spec

    for disallowed in ("pip",):
        if find_spec(disallowed):
            raise RuntimeError(f"Should not be able to import {disallowed!r}!")

    if EXPECT_CONSUMER_IMPORT:
        result = dynlib_consumer.checkdynlib_sum(1, 2)
        if result != 3:
            raise RuntimeError(f"Expected 1+2=3, got {result}")

    print("Environment launch module executed successfully")



"""Test for package index access configuration."""

import os

from pathlib import Path

import pytest

from venvstacks.stacks import PackageIndexConfig


class TestDefaultOptions:
    TEST_CONFIG = PackageIndexConfig()

    def test_uv_pip_compile(self) -> None:
        assert self.TEST_CONFIG._get_uv_pip_compile_args() == [
            "--only-binary",
            ":all:",
        ]

    def test_pip_install(self) -> None:
        assert self.TEST_CONFIG._get_pip_install_args() == [
            "--only-binary",
            ":all:",
        ]


class TestConfiguredOptions:
    TEST_CONFIG = PackageIndexConfig(
        query_default_index=False,
        local_wheel_dirs=["/some_dir"],
    )
    WHEEL_DIR = f"{os.sep}some_dir"

    def test_uv_pip_compile(self) -> None:
        # There are currently no locking specific args
        assert self.TEST_CONFIG._get_uv_pip_compile_args() == [
            "--only-binary",
            ":all:",
            "--no-index",
            "--find-links",
            self.WHEEL_DIR,
        ]

    def test_pip_install(self) -> None:
        # There are currently no installation specific args
        assert self.TEST_CONFIG._get_pip_install_args() == [
            "--only-binary",
            ":all:",
            "--no-index",
            "--find-links",
            self.WHEEL_DIR,
        ]


# Miscellaneous test cases
def test_wheel_dir_not_in_sequence() -> None:
    with pytest.raises(TypeError):
        PackageIndexConfig(local_wheel_dirs="/some_dir")


def test_lexical_path_resolution() -> None:
    paths_to_resolve = [
        "/some/path",
        "/some/absolute/../path",
        "some/path",
        "some/relative/../path",
        "~/some/path",
        "~/some/user/../path",
    ]
    expected_paths = [
        Path("/some/path").absolute(),
        Path("/some/path").absolute(),
        Path("/base_path/some/path").absolute(),
        Path("/base_path/some/path").absolute(),
        Path.home() / "some/path",
        Path.home() / "some/path",
    ]
    config = PackageIndexConfig(local_wheel_dirs=paths_to_resolve)
    config.resolve_lexical_paths("/base_path")
    assert config.local_wheel_paths == expected_paths



"""Test cases for hashing utility functions."""

import hashlib
import shutil
import tempfile

import pytest

from pathlib import Path
from typing import Generator, Mapping

from venvstacks.stacks import (
    _hash_directory,
    _hash_file_contents,
    _hash_file_name_and_contents,
    _hash_strings,
)

##################################
# Hash testing helpers
##################################

_THIS_PATH = Path(__file__)
HASH_FODDER_PATH = _THIS_PATH.parent / "hash_fodder"

# Expected content hashes generated with `sha256sum` rather than Python
# Examples ensure that file names don't affect the hash, but the file contents do
# Combined hashes are checked for consistency against their initial implementation
SHA256_ALGORITHM = "sha256"
EXPECTED_CONTENT_HASHES_SHA256: Mapping[str, str] = {
    "file.txt": "84dae841773532dcc56da3a65a4c992534c385649645bf0340873da2e2ce7d6a",
    "file_duplicate.txt": "84dae841773532dcc56da3a65a4c992534c385649645bf0340873da2e2ce7d6a",
    "different_file.txt": "43691ae21f1fd9540bb5b9a6f2ab07fd5be4c2a0545231dc505a5f33a1619337",
}
EXPECTED_COMBINED_HASHES_SHA256: Mapping[str, str] = {
    "file.txt": "6f0923fd3e132d2a05f0ee67b906d10af52e5993454ef38d024820a6fa3b2a12",
    "file_duplicate.txt": "a6d265b49ae38e1539d613d194befa0984bc0d6dcad66f46082e3e532dafc0e1",
    "different_file.txt": "4a23fc5657c4c15fce3f99c22af471a7432ac96e74ea3ca2ab7720da6b892756",
}

# Expected content hashes generated with `b2sum` rather than Python
# Examples ensure that file names don't affect the hash, but the file contents do
# Combined hashes are checked for consistency against their initial implementation
BLAKE2_ALGORITHM = "blake2b"
EXPECTED_CONTENT_HASHES_BLAKE2: Mapping[str, str] = {
    "file.txt": "bf4d9de4092670662fe8985f38880ce2d1b34ee74a4a110ea6dde23903388bc4fb18b233cc5fb027a2b374731ed6cc9274e244af5605040aa59882a7d6b68b0d",
    "file_duplicate.txt": "bf4d9de4092670662fe8985f38880ce2d1b34ee74a4a110ea6dde23903388bc4fb18b233cc5fb027a2b374731ed6cc9274e244af5605040aa59882a7d6b68b0d",
    "different_file.txt": "4783a95cdf9b6d0ebc4fe0d553ed6424b0a55400d9ead89b7c5b2dff26fb210aa1f7f9f8b809e58c7f2c79b4e046eea1b52c3a19032d2b861e792814b4ad0782",
}
EXPECTED_COMBINED_HASHES_BLAKE2: Mapping[str, str] = {
    "file.txt": "6308aff6e73bac77a1cc5d7669eedee59174040c674b2304dd73236c5ea5a3881d8ed4095700d9a22f68c84a73cee71cb965a6ba84efe546707fcea09fb150f0",
    "file_duplicate.txt": "3686fb1772e99186bc7b10ca8728d0e2d0496d6eebc9b2fda8b24a9030a0e627bbc4e119b5b8edb64caab5a9acb1b6f7aa85b79c2508562f3314dd97fb51ce7c",
    "different_file.txt": "702163efe664584b88b9a28e254c655d873bebbea4b1963bc5fdc7137dd209e8d2f743f9403563fddc3b7b27aa60f4f7a6ce87f8a6bc14971a09e4594f122bf8",
}

# Default algorithm is SHA256
DEFAULT_ALGORITHM = SHA256_ALGORITHM
EXPECTED_CONTENT_HASHES_DEFAULT = EXPECTED_CONTENT_HASHES_SHA256
EXPECTED_COMBINED_HASHES_DEFAULT = EXPECTED_COMBINED_HASHES_SHA256

# Flatten the content hash mappings into 3-tuples for easier test parameterisation
ALGORITHMS_TO_EXPECTED_CONTENT_HASHES = {
    SHA256_ALGORITHM: EXPECTED_CONTENT_HASHES_SHA256,
    BLAKE2_ALGORITHM: EXPECTED_CONTENT_HASHES_BLAKE2,
}
ALGORITHMS_TO_EXPECTED_COMBINED_HASHES = {
    SHA256_ALGORITHM: EXPECTED_COMBINED_HASHES_SHA256,
    BLAKE2_ALGORITHM: EXPECTED_COMBINED_HASHES_BLAKE2,
}


def _flatten_expected_file_hashes(
    algorithms_to_hashes: Mapping[str, Mapping[str, str]],
) -> Generator[tuple[str, str, str], None, None]:
    for algorithm, expected_hashes in algorithms_to_hashes.items():
        for fname, expected_hash in expected_hashes.items():
            yield algorithm, fname, expected_hash


EXPECTED_CONTENT_HASHES = [
    *_flatten_expected_file_hashes(ALGORITHMS_TO_EXPECTED_CONTENT_HASHES)
]
EXPECTED_COMBINED_HASHES = [
    *_flatten_expected_file_hashes(ALGORITHMS_TO_EXPECTED_COMBINED_HASHES)
]

STRINGS_TO_HASH = (
    "A string",
    "Another string",
    "A string with non-ASCII characters: 🦎🐸🦎",
)
EXPECTED_STRING_ITERABLE_HASHES = {
    SHA256_ALGORITHM: "30ad29063127d5e45ade75380df47dbbcee1246c55e138b0d42419ecb3a8635a",
    BLAKE2_ALGORITHM: "cc67ce2ed4b3b8da567de50312ea33f9207b4fcd34be07476198e7e9a20422e5c2795ef3f3c955581b53bcf715dab5945b687ef3f1e8e6a8ef1bc9bf327f72e4",
}


##########################
# Test cases
##########################


class TestStringIterableHashing:
    def test_default_hash(self) -> None:
        hash_fodder = STRINGS_TO_HASH
        expected_hash = EXPECTED_STRING_ITERABLE_HASHES[DEFAULT_ALGORITHM]
        assert _hash_strings(hash_fodder) == f"{DEFAULT_ALGORITHM}:{expected_hash}"
        assert _hash_strings(hash_fodder, omit_prefix=True) == expected_hash

    @pytest.mark.parametrize("algorithm", ALGORITHMS_TO_EXPECTED_CONTENT_HASHES)
    def test_algorithm_selection(self, algorithm: str) -> None:
        hash_fodder = STRINGS_TO_HASH
        expected_hash = EXPECTED_STRING_ITERABLE_HASHES[algorithm]
        assert _hash_strings(hash_fodder, algorithm) == f"{algorithm}:{expected_hash}"
        assert _hash_strings(hash_fodder, algorithm, omit_prefix=True) == expected_hash


class TestFileContentHashing:
    @pytest.mark.parametrize(
        "fname,expected_hash", EXPECTED_CONTENT_HASHES_DEFAULT.items()
    )
    def test_default_hash(self, fname: str, expected_hash: str) -> None:
        file_path = HASH_FODDER_PATH / fname
        assert _hash_file_contents(file_path) == f"{DEFAULT_ALGORITHM}:{expected_hash}"
        assert _hash_file_contents(file_path, omit_prefix=True) == expected_hash

    @pytest.mark.parametrize("algorithm,fname,expected_hash", EXPECTED_CONTENT_HASHES)
    def test_algorithm_selection(
        self, algorithm: str, fname: str, expected_hash: str
    ) -> None:
        file_path = HASH_FODDER_PATH / fname
        assert (
            _hash_file_contents(file_path, algorithm) == f"{algorithm}:{expected_hash}"
        )
        assert (
            _hash_file_contents(file_path, algorithm, omit_prefix=True) == expected_hash
        )


class TestFileNameAndContentHashing:
    @pytest.mark.parametrize(
        "fname,expected_hash", EXPECTED_COMBINED_HASHES_DEFAULT.items()
    )
    def test_default_hash(self, fname: str, expected_hash: str) -> None:
        file_path = HASH_FODDER_PATH / fname
        assert (
            _hash_file_name_and_contents(file_path)
            == f"{DEFAULT_ALGORITHM}:{expected_hash}"
        )
        assert (
            _hash_file_name_and_contents(file_path, omit_prefix=True) == expected_hash
        )

    @pytest.mark.parametrize("algorithm,fname,expected_hash", EXPECTED_COMBINED_HASHES)
    def test_algorithm_selection(
        self, algorithm: str, fname: str, expected_hash: str
    ) -> None:
        file_path = HASH_FODDER_PATH / fname
        assert (
            _hash_file_name_and_contents(file_path, algorithm)
            == f"{algorithm}:{expected_hash}"
        )
        assert (
            _hash_file_name_and_contents(file_path, algorithm, omit_prefix=True)
            == expected_hash
        )


# Directory hashing uses a custom algorithm (hence the non-standard prefix separator).
# However, the expected hashes for the `hash_fodder` folder can be calculated by specifying
# the expected order that different components of the hash are added to the algorithm:
#
# * directories are visited top down in sorted order
# * directory names are added to the hash when they are visited
# * file content hashes are added to the hash in sorted order after the directory name

EXPECTED_DIR_HASH_SEQUENCE = [
    ("dirname", "hash_fodder"),
    ("filename", "different_file.txt"),
    ("contents_hash", "different_file.txt"),
    ("filename", "file.txt"),
    ("contents_hash", "file.txt"),
    ("filename", "file_duplicate.txt"),
    ("contents_hash", "file_duplicate.txt"),
    ("dirname", "folder1"),
    ("filename", "file.txt"),
    ("contents_hash", "file.txt"),
    ("dirname", "subfolder"),
    ("filename", "file.txt"),
    ("contents_hash", "file.txt"),
    ("dirname", "folder2"),
    ("filename", "file_duplicate.txt"),
    ("contents_hash", "file_duplicate.txt"),
]


def _make_expected_dir_hash(algorithm: str, content_hashes: Mapping[str, str]) -> str:
    incremental_hash = hashlib.new(algorithm)
    for component_kind, component_text in EXPECTED_DIR_HASH_SEQUENCE:
        match component_kind:
            case "dirname" | "filename":
                hash_component = component_text.encode()
            case "contents_hash":
                # Directory hashing includes the algorithm prefix (at least for now)
                hash_component = (
                    f"{algorithm}:{content_hashes[component_text]}".encode()
                )
        # print(component_text, hash_component)
        incremental_hash.update(hash_component)
    return incremental_hash.hexdigest()


EXPECTED_DIR_HASHES = {
    algorithm: _make_expected_dir_hash(algorithm, expected_content_hashes)
    for algorithm, expected_content_hashes in ALGORITHMS_TO_EXPECTED_CONTENT_HASHES.items()
}


@pytest.fixture
def cloned_dir_path() -> Generator[Path, None, None]:
    with tempfile.TemporaryDirectory() as dir_name:
        temp_dir_path = Path(dir_name)
        cloned_hash_fodder_path = temp_dir_path / HASH_FODDER_PATH.name
        shutil.copytree(HASH_FODDER_PATH, cloned_hash_fodder_path)
        yield cloned_hash_fodder_path


class TestDirectoryHashing:
    def test_default_hash(self) -> None:
        dir_path = HASH_FODDER_PATH
        expected_hash = EXPECTED_DIR_HASHES[DEFAULT_ALGORITHM]
        assert _hash_directory(dir_path) == f"{DEFAULT_ALGORITHM}/{expected_hash}"
        assert _hash_directory(dir_path, omit_prefix=True) == expected_hash

    @pytest.mark.parametrize("algorithm,expected_hash", EXPECTED_DIR_HASHES.items())
    def test_algorithm_selection(self, algorithm: str, expected_hash: str) -> None:
        dir_path = HASH_FODDER_PATH
        assert _hash_directory(dir_path, algorithm) == f"{algorithm}/{expected_hash}"
        assert _hash_directory(dir_path, algorithm, omit_prefix=True) == expected_hash

    def test_root_dir_name_change_detected(self, cloned_dir_path: Path) -> None:
        renamed_dir_path = cloned_dir_path.with_name("something_completely_different")
        cloned_dir_path.rename(renamed_dir_path)
        unmodified_hash = EXPECTED_DIR_HASHES[DEFAULT_ALGORITHM]
        assert _hash_directory(renamed_dir_path, omit_prefix=True) != unmodified_hash

    def test_subdir_name_change_detected(self, cloned_dir_path: Path) -> None:
        subfolder_path = cloned_dir_path / "folder1"
        renamed_dir_path = subfolder_path.with_name("something_completely_different")
        subfolder_path.rename(renamed_dir_path)
        unmodified_hash = EXPECTED_DIR_HASHES[DEFAULT_ALGORITHM]
        assert _hash_directory(cloned_dir_path, omit_prefix=True) != unmodified_hash

    def test_file_name_change_detected(self, cloned_dir_path: Path) -> None:
        file_path = cloned_dir_path / "folder1/subfolder/file.txt"
        renamed_file_path = file_path.with_name("something_completely_different")
        file_path.rename(renamed_file_path)
        unmodified_hash = EXPECTED_DIR_HASHES[DEFAULT_ALGORITHM]
        assert _hash_directory(cloned_dir_path, omit_prefix=True) != unmodified_hash

    def test_file_contents_change_detected(self, cloned_dir_path: Path) -> None:
        file_path = cloned_dir_path / "folder1/subfolder/file.txt"
        file_path.write_text("This changes the directory hash")
        unmodified_hash = EXPECTED_DIR_HASHES[DEFAULT_ALGORITHM]
        assert _hash_directory(cloned_dir_path, omit_prefix=True) != unmodified_hash
